try:
    import cv2
except:
    raise Exception('Failed to import cv2 from the OpenCV distribution. Please install OpenCV with Python support. OpenCV may either be installed from http://opencv.org or installed with package managers such as Homebrew (on Mac, http://brew.sh) or apt-get (on Ubuntu or Debian).')


from renderer import DepthRenderer
import numpy as np
import unittest


visualize = False

class TestDepthRenderer(unittest.TestCase):

    def setUp(self):
        np.random.seed(0)

    def test_depth_image(self):
        # Create renderer
        import chumpy as ch
        from renderer import DepthRenderer
        rn = DepthRenderer()

        # Assign attributes to renderer
        from util_tests import get_earthmesh
        m = get_earthmesh(trans=ch.array([0,0,0]), rotation=ch.zeros(3))
        m.v = m.v * .01
        m.v[:,2] += 4
        w, h = (320, 240)
        from camera import ProjectPoints
        rn.camera = ProjectPoints(v=m.v, rt=ch.zeros(3), t=ch.zeros(3), f=ch.array([w,w])/2., c=ch.array([w,h])/2., k=ch.zeros(5))
        rn.frustum = {'near': 1., 'far': 10., 'width': w, 'height': h}
        rn.set(v=m.v, f=m.f, vc=m.vc*0+1, bgcolor=ch.zeros(3))
        
        # import time
        # tm = time.time()
        # rn.r
        # print 'took %es' % (time.time() - tm)

        # print np.min(rn.r.ravel())
        # print np.max(rn.r.ravel())
        self.assertLess(np.abs(np.min(rn.r.ravel()) - 3.98), 1e-5)
        self.assertLess(np.abs(np.min(m.v[:,2]) - np.min(rn.r.ravel())), 1e-5)
        self.assertLess(np.abs(rn.r[h/2,w/2] - 3.98), 1e-5)

    def test_derivatives(self):
        import chumpy as ch
        from chumpy.utils import row
        import numpy as np
        from renderer import DepthRenderer

        rn = DepthRenderer()

        # Assign attributes to renderer
        from util_tests import get_earthmesh
        m = get_earthmesh(trans=ch.array([0,0,4]), rotation=ch.zeros(3))
        w, h = (320, 240)
        from camera import ProjectPoints
        rn.camera = ProjectPoints(v=m.v, rt=ch.zeros(3), t=ch.zeros(3), f=ch.array([w,w])/2., c=ch.array([w,h])/2., k=ch.zeros(5))
        rn.frustum = {'near': 1., 'far': 10., 'width': w, 'height': h}
        rn.set(v=m.v, f=m.f, bgcolor=ch.zeros(3))

        if visualize:
            import matplotlib.pyplot as plt
            plt.figure()
        for which in range(3):
            r1 = rn.r

            adder = np.zeros(3)
            adder[which] = .01
            change = rn.v.r * 0 + row(adder)
            dr_pred = rn.dr_wrt(rn.v).dot(change.ravel()).reshape(rn.shape)
            rn.v = rn.v.r + change

            r2 = rn.r
            dr_emp = r2 - r1

            # print np.mean(np.abs(dr_pred-dr_emp))

            self.assertLess(np.mean(np.abs(dr_pred-dr_emp)), .026)

            if visualize:
                plt.subplot(2,3,which+1)
                plt.imshow(dr_pred)
                plt.clim(-.01,.01)
                plt.title('emp')
                plt.subplot(2,3,which+4)
                plt.imshow(dr_emp)
                plt.clim(-.01,.01)
                plt.title('pred')

    def test_derivatives2(self):
        import chumpy as ch
        import numpy as np
        from renderer import DepthRenderer

        rn = DepthRenderer()

        # Assign attributes to renderer
        from util_tests import get_earthmesh
        m = get_earthmesh(trans=ch.array([0,0,4]), rotation=ch.zeros(3))
        w, h = (320, 240)
        from camera import ProjectPoints
        rn.camera = ProjectPoints(v=m.v, rt=ch.zeros(3), t=ch.zeros(3), f=ch.array([w,w])/2., c=ch.array([w,h])/2., k=ch.zeros(5))
        rn.frustum = {'near': 1., 'far': 10., 'width': w, 'height': h}
        rn.set(v=m.v, f=m.f, bgcolor=ch.zeros(3))

        if visualize:
            import matplotlib.pyplot as plt
            plt.ion()
            plt.figure()

        for which in range(3):
            r1 = rn.r

            adder = np.random.rand(rn.v.r.size).reshape(rn.v.r.shape)*.01
            change = rn.v.r * 0 + adder
            dr_pred = rn.dr_wrt(rn.v).dot(change.ravel()).reshape(rn.shape)
            rn.v = rn.v.r + change

            r2 = rn.r
            dr_emp = r2 - r1

            #print np.mean(np.abs(dr_pred-dr_emp))

            self.assertLess(np.mean(np.abs(dr_pred-dr_emp)), .021)

            if visualize:
                plt.subplot(2,3,which+1)
                plt.imshow(dr_pred)
                plt.clim(-.01,.01)
                plt.title('emp')
                plt.subplot(2,3,which+4)
                plt.imshow(dr_emp)
                plt.clim(-.01,.01)
                plt.title('pred')
                plt.draw()
                plt.show()

if __name__ == '__main__':
    visualize = True
    #unittest.main()
    suite = unittest.TestLoader().loadTestsFromTestCase(TestDepthRenderer)
    unittest.TextTestRunner(verbosity=2).run(suite)
    import pdb; pdb.set_trace()



def fix_warnings():
    import platform
    if platform.system() == 'Darwin':
        # Get rid of various flags that cause warnings and/or errors
        import distutils.sysconfig as ds
        import string
        a = ds.get_config_vars()
        for k, v in a.items():
            try:
                if string.find(v, '-Wstrict-prototypes') != -1:
                    a[k] = string.replace(v, '-Wstrict-prototypes', '')
                if string.find(v, '-arch i386') != -1:
                    a[k] = string.replace(v, '-arch i386', '')
                if string.find(v, '-mno-fused-madd ') != -1:
                    a[k] = string.replace(v, '-mno-fused-madd', '')
            except:
                pass


#!/usr/bin/env python
# encoding: utf-8
"""
Author(s): Matthew Loper

See LICENCE.txt for licensing and contact information.
"""

from chumpy import Ch
import numpy as np
from chumpy.utils import row, col
from lighting import SphericalHarmonics
import unittest
try:
    import matplotlib.pyplot as plt
except:
    from dummy import dummy as plt
from topology import loop_subdivider

visualize = False

def getcam():
    from camera import ProjectPoints3D

    w = 640
    h = 320

    f = np.array([500,500])
    rt = np.zeros(3)
    t = np.zeros(3)
    k = np.zeros(5)
    c = np.array([w/2., h/2.])
    near = .1
    far = 20.

    frustum = {'near': near, 'far': far, 'width': w, 'height': h}
    pp = ProjectPoints3D(f=f, rt=rt, t=t, k=k, c=c)

    return pp, frustum



class TestSphericalHarmonics(unittest.TestCase):

    def test_spherical_harmonics(self):
        global visualize
        if visualize:
            plt.ion()
    
        # Get mesh
        v, f = get_sphere_mesh()

        from geometry import VertNormals
        vn = VertNormals(v=v, f=f)
        #vn =  Ch(mesh.estimate_vertex_normals())

        # Get camera
        cam, frustum = getcam()
    
        # Get renderer
        from renderer import ColoredRenderer
        cam.v = v
        cr = ColoredRenderer(f=f, camera=cam, frustum=frustum, v=v)
    
        sh_red = SphericalHarmonics(vn=vn, light_color=np.array([1,0,0]))
        sh_green = SphericalHarmonics(vn=vn, light_color=np.array([0,1,0]))
    
        cr.vc = sh_red + sh_green
        
        ims_baseline = []
        for comp_idx, subplot_idx in enumerate([3,7,8,9,11,12,13,14,15]):
        
            sh_comps = np.zeros(9)
            sh_comps[comp_idx] = 1
            sh_red.components =  Ch(sh_comps)
            sh_green.components =  Ch(-sh_comps)
            
            newim = cr.r.reshape((frustum['height'], frustum['width'], 3))
            ims_baseline.append(newim)

            if visualize:
                plt.subplot(3,5,subplot_idx)
                plt.imshow(newim)
                plt.axis('off')
            
        offset = row(.4 * (np.random.rand(3)-.5))
        #offset = row(np.array([1.,1.,1.]))*.05
        vn_shifted = (vn.r + offset)
        vn_shifted = vn_shifted / col(np.sqrt(np.sum(vn_shifted**2, axis=1)))
        vn_shifted = vn_shifted.ravel()
        vn_shifted[vn_shifted>1.] = 1
        vn_shifted[vn_shifted<-1.] = -1
        vn_shifted = Ch(vn_shifted)
        cr.replace(sh_red.vn, vn_shifted)
        if True:
            for comp_idx in range(9):
                if visualize:
                    plt.figure(comp_idx+2)
        
                sh_comps = np.zeros(9)
                sh_comps[comp_idx] = 1
                sh_red.components =  Ch(sh_comps)
                sh_green.components =  Ch(-sh_comps)
        
                pred = cr.dr_wrt(vn_shifted).dot(col(vn_shifted.r.reshape(vn.r.shape) - vn.r)).reshape((frustum['height'], frustum['width'], 3))
                if visualize:
                    plt.subplot(1,2,1)
                    plt.imshow(pred)
                    plt.title('pred (comp %d)' % (comp_idx,))        
                    plt.subplot(1,2,2)
                    
                newim = cr.r.reshape((frustum['height'], frustum['width'], 3))
                emp = newim - ims_baseline[comp_idx]
                if visualize:
                    plt.imshow(emp)
                    plt.title('empirical (comp %d)' % (comp_idx,))
                pred_flat = pred.ravel()
                emp_flat = emp.ravel()
                nnz = np.unique(np.concatenate((np.nonzero(pred_flat)[0], np.nonzero(emp_flat)[0])))
                
                if comp_idx != 0:
                    med_diff = np.median(np.abs(pred_flat[nnz]-emp_flat[nnz]))
                    med_obs = np.median(np.abs(emp_flat[nnz]))
                    if comp_idx == 4 or comp_idx == 8:
                        self.assertTrue(med_diff / med_obs < .6)
                    else:
                        self.assertTrue(med_diff / med_obs < .3)
                if visualize:
                    plt.axis('off')
    

def get_sphere_mesh():
    from util_tests import get_earthmesh

    mesh = get_earthmesh(np.zeros(3), np.zeros(3)) # load_mesh(filename)
    v, f = mesh.v*64., mesh.f

    for i in range(3):
        mtx, f = loop_subdivider(v, f)
        v = mtx.dot(v.ravel()).reshape((-1,3))
    v /= 200.
    v[:,2] += 2

    return v, f




if __name__ == '__main__':
    visualize = True
    plt.ion()
    #unittest.main()
    suite = unittest.TestLoader().loadTestsFromTestCase(TestSphericalHarmonics)
    unittest.TextTestRunner(verbosity=2).run(suite)
    plt.show()
    import pdb; pdb.set_trace()




from opendr.camera import ProjectPoints


import unittest


visualize = False

class TestOcclusion(unittest.TestCase):


    def test_occlusion(self):
        if visualize:
            import matplotlib.pyplot as plt
            plt.ion()

        # Create renderer
        import chumpy as ch
        import numpy as np
        from opendr.renderer import TexturedRenderer, ColoredRenderer
        #rn = TexturedRenderer()
        rn = ColoredRenderer()

        # Assign attributes to renderer
        from util_tests import get_earthmesh
        m = get_earthmesh(trans=ch.array([0,0,4]), rotation=ch.zeros(3))
        rn.texture_image = m.texture_image
        rn.ft = m.ft
        rn.vt = m.vt
        m.v[:,2] = np.mean(m.v[:,2])

        # red is front and zero
        # green is back and 1
        t0 = ch.array([1,0,.1])
        t1 = ch.array([-1,0,.1])
        v0 = ch.array(m.v) + t0

        if False:
            v1 = ch.array(m.v*.4 + np.array([0,0,3.8])) + t1
        else:
            v1 = ch.array(m.v) + t1
        vc0 = v0*0 + np.array([[.4,0,0]])
        vc1 = v1*0 + np.array([[0,.4,0]])
        vc = ch.vstack((vc0, vc1))

        v = ch.vstack((v0, v1))
        f = np.vstack((m.f, m.f+len(v0)))

        w, h = (320, 240)
        rn.camera = ProjectPoints(v=v, rt=ch.zeros(3), t=ch.zeros(3), f=ch.array([w,w])/2., c=ch.array([w,h])/2., k=ch.zeros(5))
        rn.camera.t = ch.array([0,0,-2.5])
        rn.frustum = {'near': 1., 'far': 10., 'width': w, 'height': h}
        m.vc = v.r*0 + np.array([[1,0,0]])
        rn.set(v=v, f=f, vc=vc)

        t0[:] = np.array([1.4, 0, .1-.02])
        t1[:] = np.array([-0.6, 0, .1+.02])

        target = rn.r

        if visualize:
            plt.figure()
            plt.imshow(target)
            plt.title('target')

            plt.figure()
            plt.show()

        im_orig = rn.r.copy()

        from cvwrap import cv2

        tr = t0
        eps_emp = .02
        eps_pred = .02

        #blur = lambda x : cv2.blur(x, ksize=(5,5))
        blur = lambda x : x
        for tr in [t0, t1]:
            if tr is t0:
                sum_limits = np.array([2.1e+2, 6.9e+1, 1.6e+2])
            else:
                sum_limits = [1., 5., 4.]

            if visualize:
                plt.figure()
            for i in range(3):
                dr_pred = np.array(rn.dr_wrt(tr[i]).todense()).reshape(rn.shape) * eps_pred
                dr_pred = blur(dr_pred)

                # central differences
                tr[i] = tr[i].r + eps_emp/2.
                rn_greater = rn.r.copy()
                tr[i] = tr[i].r - eps_emp/1.
                rn_lesser = rn.r.copy()
                tr[i] = tr[i].r + eps_emp/2.

                dr_emp = blur((rn_greater - rn_lesser) * eps_pred / eps_emp)

                dr_pred_shown = np.clip(dr_pred, -.5, .5) + .5
                dr_emp_shown = np.clip(dr_emp, -.5, .5) + .5

                if visualize:
                    plt.subplot(3,3,i+1)
                    plt.imshow(dr_pred_shown)
                    plt.title('pred')
                    plt.axis('off')

                    plt.subplot(3,3,3+i+1)
                    plt.imshow(dr_emp_shown)
                    plt.title('empirical')
                    plt.axis('off')

                    plt.subplot(3,3,6+i+1)

                diff = np.abs(dr_emp - dr_pred)
                if visualize:
                    plt.imshow(diff)
                diff = diff.ravel()
                if visualize:
                    plt.title('diff (sum: %.2e)'  % (np.sum(diff)))
                    plt.axis('off')

                # print 'dr pred sum: %.2e' % (np.sum(np.abs(dr_pred.ravel())),)
                # print 'dr emp sum: %.2e' % (np.sum(np.abs(dr_emp.ravel())),)

                #import pdb; pdb.set_trace()
                self.assertTrue(np.sum(diff) < sum_limits[i])







if __name__ == '__main__':
    visualize = True
    suite = unittest.TestLoader().loadTestsFromTestCase(TestOcclusion)
    unittest.TextTestRunner(verbosity=2).run(suite)
    import pdb; pdb.set_trace()



cdef extern from "gl_includes.h":
	ctypedef unsigned int	GLenum
	ctypedef unsigned char	GLboolean
	ctypedef unsigned int	GLbitfield
	ctypedef void		GLvoid
	ctypedef signed char	GLbyte
	ctypedef short		GLshort
	ctypedef int		GLint
	ctypedef unsigned char	GLubyte
	ctypedef unsigned short	GLushort
	ctypedef unsigned int	GLuint
	ctypedef int		GLsizei
	ctypedef float		GLfloat
	ctypedef float		GLclampf
	ctypedef double		GLdouble
	ctypedef double		GLclampd
	ctypedef void* GLeglImageOES
	ctypedef char GLchar
	ctypedef ptrdiff_t GLintptr
	ctypedef ptrdiff_t GLsizeiptr
	cdef void _glClearIndex "glClearIndex"( GLfloat c ) nogil
	cdef void _glClearColor "glClearColor"( GLclampf red, GLclampf green, GLclampf blue, GLclampf alpha ) nogil
	cdef void _glClear "glClear"( GLbitfield mask ) nogil
	cdef void _glIndexMask "glIndexMask"( GLuint mask ) nogil
	cdef void _glColorMask "glColorMask"( GLboolean red, GLboolean green, GLboolean blue, GLboolean alpha ) nogil
	cdef void _glAlphaFunc "glAlphaFunc"( GLenum func, GLclampf ref ) nogil
	cdef void _glBlendFunc "glBlendFunc"( GLenum sfactor, GLenum dfactor ) nogil
	cdef void _glLogicOp "glLogicOp"( GLenum opcode ) nogil
	cdef void _glCullFace "glCullFace"( GLenum mode ) nogil
	cdef void _glFrontFace "glFrontFace"( GLenum mode ) nogil
	cdef void _glPointSize "glPointSize"( GLfloat size ) nogil
	cdef void _glLineWidth "glLineWidth"( GLfloat width ) nogil
	cdef void _glLineStipple "glLineStipple"( GLint factor, GLushort pattern ) nogil
	cdef void _glPolygonMode "glPolygonMode"( GLenum face, GLenum mode ) nogil
	cdef void _glPolygonOffset "glPolygonOffset"( GLfloat factor, GLfloat units ) nogil
	cdef void _glPolygonStipple "glPolygonStipple"(  GLubyte *mask ) 
	cdef void _glGetPolygonStipple "glGetPolygonStipple"( GLubyte *mask ) 
	cdef void _glEdgeFlag "glEdgeFlag"( GLboolean flag ) nogil
	cdef void _glEdgeFlagv "glEdgeFlagv"(  GLboolean *flag ) 
	cdef void _glScissor "glScissor"( GLint x, GLint y, GLsizei width, GLsizei height) nogil
	cdef void _glClipPlane "glClipPlane"( GLenum plane,  GLdouble *equation ) 
	cdef void _glGetClipPlane "glGetClipPlane"( GLenum plane, GLdouble *equation ) 
	cdef void _glDrawBuffer "glDrawBuffer"( GLenum mode ) nogil
	cdef void _glReadBuffer "glReadBuffer"( GLenum mode ) nogil
	cdef void _glEnable "glEnable"( GLenum cap ) nogil
	cdef void _glDisable "glDisable"( GLenum cap ) nogil
	cdef GLboolean _glIsEnabled "glIsEnabled"( GLenum cap ) nogil
	cdef void _glEnableClientState "glEnableClientState"( GLenum cap ) nogil
	cdef void _glDisableClientState "glDisableClientState"( GLenum cap ) nogil
	cdef void _glGetBooleanv "glGetBooleanv"( GLenum pname, GLboolean *params ) 
	cdef void _glGetDoublev "glGetDoublev"( GLenum pname, GLdouble *params ) 
	cdef void _glGetFloatv "glGetFloatv"( GLenum pname, GLfloat *params ) 
	cdef void _glGetIntegerv "glGetIntegerv"( GLenum pname, GLint *params ) 
	cdef void _glPushAttrib "glPushAttrib"( GLbitfield mask ) nogil
	cdef void _glPopAttrib "glPopAttrib"() nogil
	cdef void _glPushClientAttrib "glPushClientAttrib"( GLbitfield mask ) nogil
	cdef void _glPopClientAttrib "glPopClientAttrib"() nogil
	cdef GLint _glRenderMode "glRenderMode"( GLenum mode ) nogil
	cdef GLenum _glGetError "glGetError"() nogil
	cdef void _glFinish "glFinish"() nogil
	cdef void _glFlush "glFlush"() nogil
	cdef void _glHint "glHint"( GLenum target, GLenum mode ) nogil
	cdef void _glClearDepth "glClearDepth"( GLclampd depth ) nogil
	cdef void _glDepthFunc "glDepthFunc"( GLenum func ) nogil
	cdef void _glDepthMask "glDepthMask"( GLboolean flag ) nogil
	cdef void _glDepthRange "glDepthRange"( GLclampd near_val, GLclampd far_val ) nogil
	cdef void _glClearAccum "glClearAccum"( GLfloat red, GLfloat green, GLfloat blue, GLfloat alpha ) nogil
	cdef void _glAccum "glAccum"( GLenum op, GLfloat value ) nogil
	cdef void _glMatrixMode "glMatrixMode"( GLenum mode ) nogil
	cdef void _glOrtho "glOrtho"( GLdouble left, GLdouble right,
                                 GLdouble bottom, GLdouble top,
                                 GLdouble near_val, GLdouble far_val ) nogil
	cdef void _glFrustum "glFrustum"( GLdouble left, GLdouble right,
                                   GLdouble bottom, GLdouble top,
                                   GLdouble near_val, GLdouble far_val ) nogil
	cdef void _glViewport "glViewport"( GLint x, GLint y,
                                    GLsizei width, GLsizei height ) nogil
	cdef void _glPushMatrix "glPushMatrix"() nogil
	cdef void _glPopMatrix "glPopMatrix"() nogil
	cdef void _glLoadIdentity "glLoadIdentity"() nogil
	cdef void _glLoadMatrixd "glLoadMatrixd"(  GLdouble *m ) 
	cdef void _glLoadMatrixf "glLoadMatrixf"(  GLfloat *m ) 
	cdef void _glMultMatrixd "glMultMatrixd"(  GLdouble *m ) 
	cdef void _glMultMatrixf "glMultMatrixf"(  GLfloat *m ) 
	cdef void _glRotated "glRotated"( GLdouble angle,
                                   GLdouble x, GLdouble y, GLdouble z ) nogil
	cdef void _glRotatef "glRotatef"( GLfloat angle,
                                   GLfloat x, GLfloat y, GLfloat z ) nogil
	cdef void _glScaled "glScaled"( GLdouble x, GLdouble y, GLdouble z ) nogil
	cdef void _glScalef "glScalef"( GLfloat x, GLfloat y, GLfloat z ) nogil
	cdef void _glTranslated "glTranslated"( GLdouble x, GLdouble y, GLdouble z ) nogil
	cdef void _glTranslatef "glTranslatef"( GLfloat x, GLfloat y, GLfloat z ) nogil
	cdef GLboolean _glIsList "glIsList"( GLuint list ) nogil
	cdef void _glDeleteLists "glDeleteLists"( GLuint list, GLsizei range ) nogil
	cdef GLuint _glGenLists "glGenLists"( GLsizei range ) nogil
	cdef void _glNewList "glNewList"( GLuint list, GLenum mode ) nogil
	cdef void _glEndList "glEndList"() nogil
	cdef void _glCallList "glCallList"( GLuint list ) nogil
	cdef void _glCallLists "glCallLists"( GLsizei n, GLenum type,
                                      GLvoid *lists ) 
	cdef void _glListBase "glListBase"( GLuint base ) nogil
	cdef void _glBegin "glBegin"( GLenum mode ) nogil
	cdef void _glEnd "glEnd"() nogil
	cdef void _glVertex2d "glVertex2d"( GLdouble x, GLdouble y ) nogil
	cdef void _glVertex2f "glVertex2f"( GLfloat x, GLfloat y ) nogil
	cdef void _glVertex2i "glVertex2i"( GLint x, GLint y ) nogil
	cdef void _glVertex2s "glVertex2s"( GLshort x, GLshort y ) nogil
	cdef void _glVertex3d "glVertex3d"( GLdouble x, GLdouble y, GLdouble z ) nogil
	cdef void _glVertex3f "glVertex3f"( GLfloat x, GLfloat y, GLfloat z ) nogil
	cdef void _glVertex3i "glVertex3i"( GLint x, GLint y, GLint z ) nogil
	cdef void _glVertex3s "glVertex3s"( GLshort x, GLshort y, GLshort z ) nogil
	cdef void _glVertex4d "glVertex4d"( GLdouble x, GLdouble y, GLdouble z, GLdouble w ) nogil
	cdef void _glVertex4f "glVertex4f"( GLfloat x, GLfloat y, GLfloat z, GLfloat w ) nogil
	cdef void _glVertex4i "glVertex4i"( GLint x, GLint y, GLint z, GLint w ) nogil
	cdef void _glVertex4s "glVertex4s"( GLshort x, GLshort y, GLshort z, GLshort w ) nogil
	cdef void _glVertex2dv "glVertex2dv"(  GLdouble *v ) 
	cdef void _glVertex2fv "glVertex2fv"(  GLfloat *v ) 
	cdef void _glVertex2iv "glVertex2iv"(  GLint *v ) 
	cdef void _glVertex2sv "glVertex2sv"(  GLshort *v ) 
	cdef void _glVertex3dv "glVertex3dv"(  GLdouble *v ) 
	cdef void _glVertex3fv "glVertex3fv"(  GLfloat *v ) 
	cdef void _glVertex3iv "glVertex3iv"(  GLint *v ) 
	cdef void _glVertex3sv "glVertex3sv"(  GLshort *v ) 
	cdef void _glVertex4dv "glVertex4dv"(  GLdouble *v ) 
	cdef void _glVertex4fv "glVertex4fv"(  GLfloat *v ) 
	cdef void _glVertex4iv "glVertex4iv"(  GLint *v ) 
	cdef void _glVertex4sv "glVertex4sv"(  GLshort *v ) 
	cdef void _glNormal3b "glNormal3b"( GLbyte nx, GLbyte ny, GLbyte nz ) nogil
	cdef void _glNormal3d "glNormal3d"( GLdouble nx, GLdouble ny, GLdouble nz ) nogil
	cdef void _glNormal3f "glNormal3f"( GLfloat nx, GLfloat ny, GLfloat nz ) nogil
	cdef void _glNormal3i "glNormal3i"( GLint nx, GLint ny, GLint nz ) nogil
	cdef void _glNormal3s "glNormal3s"( GLshort nx, GLshort ny, GLshort nz ) nogil
	cdef void _glNormal3bv "glNormal3bv"(  GLbyte *v ) 
	cdef void _glNormal3dv "glNormal3dv"(  GLdouble *v ) 
	cdef void _glNormal3fv "glNormal3fv"(  GLfloat *v ) 
	cdef void _glNormal3iv "glNormal3iv"(  GLint *v ) 
	cdef void _glNormal3sv "glNormal3sv"(  GLshort *v ) 
	cdef void _glIndexd "glIndexd"( GLdouble c ) nogil
	cdef void _glIndexf "glIndexf"( GLfloat c ) nogil
	cdef void _glIndexi "glIndexi"( GLint c ) nogil
	cdef void _glIndexs "glIndexs"( GLshort c ) nogil
	cdef void _glIndexub "glIndexub"( GLubyte c ) nogil
	cdef void _glIndexdv "glIndexdv"(  GLdouble *c ) 
	cdef void _glIndexfv "glIndexfv"(  GLfloat *c ) 
	cdef void _glIndexiv "glIndexiv"(  GLint *c ) 
	cdef void _glIndexsv "glIndexsv"(  GLshort *c ) 
	cdef void _glIndexubv "glIndexubv"(  GLubyte *c ) 
	cdef void _glColor3b "glColor3b"( GLbyte red, GLbyte green, GLbyte blue ) nogil
	cdef void _glColor3d "glColor3d"( GLdouble red, GLdouble green, GLdouble blue ) nogil
	cdef void _glColor3f "glColor3f"( GLfloat red, GLfloat green, GLfloat blue ) nogil
	cdef void _glColor3i "glColor3i"( GLint red, GLint green, GLint blue ) nogil
	cdef void _glColor3s "glColor3s"( GLshort red, GLshort green, GLshort blue ) nogil
	cdef void _glColor3ub "glColor3ub"( GLubyte red, GLubyte green, GLubyte blue ) nogil
	cdef void _glColor3ui "glColor3ui"( GLuint red, GLuint green, GLuint blue ) nogil
	cdef void _glColor3us "glColor3us"( GLushort red, GLushort green, GLushort blue ) nogil
	cdef void _glColor4b "glColor4b"( GLbyte red, GLbyte green,
                                   GLbyte blue, GLbyte alpha ) nogil
	cdef void _glColor4d "glColor4d"( GLdouble red, GLdouble green,
                                   GLdouble blue, GLdouble alpha ) nogil
	cdef void _glColor4f "glColor4f"( GLfloat red, GLfloat green,
                                   GLfloat blue, GLfloat alpha ) nogil
	cdef void _glColor4i "glColor4i"( GLint red, GLint green,
                                   GLint blue, GLint alpha ) nogil
	cdef void _glColor4s "glColor4s"( GLshort red, GLshort green,
                                   GLshort blue, GLshort alpha ) nogil
	cdef void _glColor4ub "glColor4ub"( GLubyte red, GLubyte green,
                                    GLubyte blue, GLubyte alpha ) nogil
	cdef void _glColor4ui "glColor4ui"( GLuint red, GLuint green,
                                    GLuint blue, GLuint alpha ) nogil
	cdef void _glColor4us "glColor4us"( GLushort red, GLushort green,
                                    GLushort blue, GLushort alpha ) nogil
	cdef void _glColor3bv "glColor3bv"(  GLbyte *v ) 
	cdef void _glColor3dv "glColor3dv"(  GLdouble *v ) 
	cdef void _glColor3fv "glColor3fv"(  GLfloat *v ) 
	cdef void _glColor3iv "glColor3iv"(  GLint *v ) 
	cdef void _glColor3sv "glColor3sv"(  GLshort *v ) 
	cdef void _glColor3ubv "glColor3ubv"(  GLubyte *v ) 
	cdef void _glColor3uiv "glColor3uiv"(  GLuint *v ) 
	cdef void _glColor3usv "glColor3usv"(  GLushort *v ) 
	cdef void _glColor4bv "glColor4bv"(  GLbyte *v ) 
	cdef void _glColor4dv "glColor4dv"(  GLdouble *v ) 
	cdef void _glColor4fv "glColor4fv"(  GLfloat *v ) 
	cdef void _glColor4iv "glColor4iv"(  GLint *v ) 
	cdef void _glColor4sv "glColor4sv"(  GLshort *v ) 
	cdef void _glColor4ubv "glColor4ubv"(  GLubyte *v ) 
	cdef void _glColor4uiv "glColor4uiv"(  GLuint *v ) 
	cdef void _glColor4usv "glColor4usv"(  GLushort *v ) 
	cdef void _glTexCoord1d "glTexCoord1d"( GLdouble s ) nogil
	cdef void _glTexCoord1f "glTexCoord1f"( GLfloat s ) nogil
	cdef void _glTexCoord1i "glTexCoord1i"( GLint s ) nogil
	cdef void _glTexCoord1s "glTexCoord1s"( GLshort s ) nogil
	cdef void _glTexCoord2d "glTexCoord2d"( GLdouble s, GLdouble t ) nogil
	cdef void _glTexCoord2f "glTexCoord2f"( GLfloat s, GLfloat t ) nogil
	cdef void _glTexCoord2i "glTexCoord2i"( GLint s, GLint t ) nogil
	cdef void _glTexCoord2s "glTexCoord2s"( GLshort s, GLshort t ) nogil
	cdef void _glTexCoord3d "glTexCoord3d"( GLdouble s, GLdouble t, GLdouble r ) nogil
	cdef void _glTexCoord3f "glTexCoord3f"( GLfloat s, GLfloat t, GLfloat r ) nogil
	cdef void _glTexCoord3i "glTexCoord3i"( GLint s, GLint t, GLint r ) nogil
	cdef void _glTexCoord3s "glTexCoord3s"( GLshort s, GLshort t, GLshort r ) nogil
	cdef void _glTexCoord4d "glTexCoord4d"( GLdouble s, GLdouble t, GLdouble r, GLdouble q ) nogil
	cdef void _glTexCoord4f "glTexCoord4f"( GLfloat s, GLfloat t, GLfloat r, GLfloat q ) nogil
	cdef void _glTexCoord4i "glTexCoord4i"( GLint s, GLint t, GLint r, GLint q ) nogil
	cdef void _glTexCoord4s "glTexCoord4s"( GLshort s, GLshort t, GLshort r, GLshort q ) nogil
	cdef void _glTexCoord1dv "glTexCoord1dv"(  GLdouble *v ) 
	cdef void _glTexCoord1fv "glTexCoord1fv"(  GLfloat *v ) 
	cdef void _glTexCoord1iv "glTexCoord1iv"(  GLint *v ) 
	cdef void _glTexCoord1sv "glTexCoord1sv"(  GLshort *v ) 
	cdef void _glTexCoord2dv "glTexCoord2dv"(  GLdouble *v ) 
	cdef void _glTexCoord2fv "glTexCoord2fv"(  GLfloat *v ) 
	cdef void _glTexCoord2iv "glTexCoord2iv"(  GLint *v ) 
	cdef void _glTexCoord2sv "glTexCoord2sv"(  GLshort *v ) 
	cdef void _glTexCoord3dv "glTexCoord3dv"(  GLdouble *v ) 
	cdef void _glTexCoord3fv "glTexCoord3fv"(  GLfloat *v ) 
	cdef void _glTexCoord3iv "glTexCoord3iv"(  GLint *v ) 
	cdef void _glTexCoord3sv "glTexCoord3sv"(  GLshort *v ) 
	cdef void _glTexCoord4dv "glTexCoord4dv"(  GLdouble *v ) 
	cdef void _glTexCoord4fv "glTexCoord4fv"(  GLfloat *v ) 
	cdef void _glTexCoord4iv "glTexCoord4iv"(  GLint *v ) 
	cdef void _glTexCoord4sv "glTexCoord4sv"(  GLshort *v ) 
	cdef void _glRasterPos2d "glRasterPos2d"( GLdouble x, GLdouble y ) nogil
	cdef void _glRasterPos2f "glRasterPos2f"( GLfloat x, GLfloat y ) nogil
	cdef void _glRasterPos2i "glRasterPos2i"( GLint x, GLint y ) nogil
	cdef void _glRasterPos2s "glRasterPos2s"( GLshort x, GLshort y ) nogil
	cdef void _glRasterPos3d "glRasterPos3d"( GLdouble x, GLdouble y, GLdouble z ) nogil
	cdef void _glRasterPos3f "glRasterPos3f"( GLfloat x, GLfloat y, GLfloat z ) nogil
	cdef void _glRasterPos3i "glRasterPos3i"( GLint x, GLint y, GLint z ) nogil
	cdef void _glRasterPos3s "glRasterPos3s"( GLshort x, GLshort y, GLshort z ) nogil
	cdef void _glRasterPos4d "glRasterPos4d"( GLdouble x, GLdouble y, GLdouble z, GLdouble w ) nogil
	cdef void _glRasterPos4f "glRasterPos4f"( GLfloat x, GLfloat y, GLfloat z, GLfloat w ) nogil
	cdef void _glRasterPos4i "glRasterPos4i"( GLint x, GLint y, GLint z, GLint w ) nogil
	cdef void _glRasterPos4s "glRasterPos4s"( GLshort x, GLshort y, GLshort z, GLshort w ) nogil
	cdef void _glRasterPos2dv "glRasterPos2dv"(  GLdouble *v ) 
	cdef void _glRasterPos2fv "glRasterPos2fv"(  GLfloat *v ) 
	cdef void _glRasterPos2iv "glRasterPos2iv"(  GLint *v ) 
	cdef void _glRasterPos2sv "glRasterPos2sv"(  GLshort *v ) 
	cdef void _glRasterPos3dv "glRasterPos3dv"(  GLdouble *v ) 
	cdef void _glRasterPos3fv "glRasterPos3fv"(  GLfloat *v ) 
	cdef void _glRasterPos3iv "glRasterPos3iv"(  GLint *v ) 
	cdef void _glRasterPos3sv "glRasterPos3sv"(  GLshort *v ) 
	cdef void _glRasterPos4dv "glRasterPos4dv"(  GLdouble *v ) 
	cdef void _glRasterPos4fv "glRasterPos4fv"(  GLfloat *v ) 
	cdef void _glRasterPos4iv "glRasterPos4iv"(  GLint *v ) 
	cdef void _glRasterPos4sv "glRasterPos4sv"(  GLshort *v ) 
	cdef void _glRectd "glRectd"( GLdouble x1, GLdouble y1, GLdouble x2, GLdouble y2 ) nogil
	cdef void _glRectf "glRectf"( GLfloat x1, GLfloat y1, GLfloat x2, GLfloat y2 ) nogil
	cdef void _glRecti "glRecti"( GLint x1, GLint y1, GLint x2, GLint y2 ) nogil
	cdef void _glRects "glRects"( GLshort x1, GLshort y1, GLshort x2, GLshort y2 ) nogil
	cdef void _glRectdv "glRectdv"(  GLdouble *v1,  GLdouble *v2 ) 
	cdef void _glRectfv "glRectfv"(  GLfloat *v1,  GLfloat *v2 ) 
	cdef void _glRectiv "glRectiv"(  GLint *v1,  GLint *v2 ) 
	cdef void _glRectsv "glRectsv"(  GLshort *v1,  GLshort *v2 ) 
	cdef void _glVertexPointer "glVertexPointer"( GLint size, GLenum type,
                                       GLsizei stride,  GLvoid *ptr ) 
	cdef void _glNormalPointer "glNormalPointer"( GLenum type, GLsizei stride,
                                        GLvoid *ptr ) 
	cdef void _glColorPointer "glColorPointer"( GLint size, GLenum type,
                                      GLsizei stride,  GLvoid *ptr ) 
	cdef void _glIndexPointer "glIndexPointer"( GLenum type, GLsizei stride,
                                       GLvoid *ptr ) 
	cdef void _glTexCoordPointer "glTexCoordPointer"( GLint size, GLenum type,
                                         GLsizei stride,  GLvoid *ptr ) 
	cdef void _glEdgeFlagPointer "glEdgeFlagPointer"( GLsizei stride,  GLvoid *ptr ) 
	cdef void _glGetPointerv "glGetPointerv"( GLenum pname, GLvoid **params ) 
	cdef void _glArrayElement "glArrayElement"( GLint i ) nogil
	cdef void _glDrawArrays "glDrawArrays"( GLenum mode, GLint first, GLsizei count ) nogil
	cdef void _glDrawElements "glDrawElements"( GLenum mode, GLsizei count,
                                      GLenum type,  GLvoid *indices ) 
	cdef void _glInterleavedArrays "glInterleavedArrays"( GLenum format, GLsizei stride,
                                            GLvoid *pointer ) 
	cdef void _glShadeModel "glShadeModel"( GLenum mode ) nogil
	cdef void _glLightf "glLightf"( GLenum light, GLenum pname, GLfloat param ) nogil
	cdef void _glLighti "glLighti"( GLenum light, GLenum pname, GLint param ) nogil
	cdef void _glLightfv "glLightfv"( GLenum light, GLenum pname,
                                  GLfloat *params ) 
	cdef void _glLightiv "glLightiv"( GLenum light, GLenum pname,
                                  GLint *params ) 
	cdef void _glGetLightfv "glGetLightfv"( GLenum light, GLenum pname,
                                    GLfloat *params ) 
	cdef void _glGetLightiv "glGetLightiv"( GLenum light, GLenum pname,
                                    GLint *params ) 
	cdef void _glLightModelf "glLightModelf"( GLenum pname, GLfloat param ) nogil
	cdef void _glLightModeli "glLightModeli"( GLenum pname, GLint param ) nogil
	cdef void _glLightModelfv "glLightModelfv"( GLenum pname,  GLfloat *params ) 
	cdef void _glLightModeliv "glLightModeliv"( GLenum pname,  GLint *params ) 
	cdef void _glMaterialf "glMaterialf"( GLenum face, GLenum pname, GLfloat param ) nogil
	cdef void _glMateriali "glMateriali"( GLenum face, GLenum pname, GLint param ) nogil
	cdef void _glMaterialfv "glMaterialfv"( GLenum face, GLenum pname,  GLfloat *params ) 
	cdef void _glMaterialiv "glMaterialiv"( GLenum face, GLenum pname,  GLint *params ) 
	cdef void _glGetMaterialfv "glGetMaterialfv"( GLenum face, GLenum pname, GLfloat *params ) 
	cdef void _glGetMaterialiv "glGetMaterialiv"( GLenum face, GLenum pname, GLint *params ) 
	cdef void _glColorMaterial "glColorMaterial"( GLenum face, GLenum mode ) nogil
	cdef void _glPixelZoom "glPixelZoom"( GLfloat xfactor, GLfloat yfactor ) nogil
	cdef void _glPixelStoref "glPixelStoref"( GLenum pname, GLfloat param ) nogil
	cdef void _glPixelStorei "glPixelStorei"( GLenum pname, GLint param ) nogil
	cdef void _glPixelTransferf "glPixelTransferf"( GLenum pname, GLfloat param ) nogil
	cdef void _glPixelTransferi "glPixelTransferi"( GLenum pname, GLint param ) nogil
	cdef void _glPixelMapfv "glPixelMapfv"( GLenum map, GLsizei mapsize,
                                     GLfloat *values ) 
	cdef void _glPixelMapuiv "glPixelMapuiv"( GLenum map, GLsizei mapsize,
                                      GLuint *values ) 
	cdef void _glPixelMapusv "glPixelMapusv"( GLenum map, GLsizei mapsize,
                                      GLushort *values ) 
	cdef void _glGetPixelMapfv "glGetPixelMapfv"( GLenum map, GLfloat *values ) 
	cdef void _glGetPixelMapuiv "glGetPixelMapuiv"( GLenum map, GLuint *values ) 
	cdef void _glGetPixelMapusv "glGetPixelMapusv"( GLenum map, GLushort *values ) 
	cdef void _glBitmap "glBitmap"( GLsizei width, GLsizei height,
                                GLfloat xorig, GLfloat yorig,
                                GLfloat xmove, GLfloat ymove,
                                 GLubyte *bitmap ) 
	cdef void _glReadPixels "glReadPixels"( GLint x, GLint y,
                                    GLsizei width, GLsizei height,
                                    GLenum format, GLenum type,
                                    GLvoid *pixels ) nogil
	cdef void _glDrawPixels "glDrawPixels"( GLsizei width, GLsizei height,
                                    GLenum format, GLenum type,
                                     GLvoid *pixels ) 
	cdef void _glCopyPixels "glCopyPixels"( GLint x, GLint y,
                                    GLsizei width, GLsizei height,
                                    GLenum type ) nogil
	cdef void _glStencilFunc "glStencilFunc"( GLenum func, GLint ref, GLuint mask ) nogil
	cdef void _glStencilMask "glStencilMask"( GLuint mask ) nogil
	cdef void _glStencilOp "glStencilOp"( GLenum fail, GLenum zfail, GLenum zpass ) nogil
	cdef void _glClearStencil "glClearStencil"( GLint s ) nogil
	cdef void _glTexGend "glTexGend"( GLenum coord, GLenum pname, GLdouble param ) nogil
	cdef void _glTexGenf "glTexGenf"( GLenum coord, GLenum pname, GLfloat param ) nogil
	cdef void _glTexGeni "glTexGeni"( GLenum coord, GLenum pname, GLint param ) nogil
	cdef void _glTexGendv "glTexGendv"( GLenum coord, GLenum pname,  GLdouble *params ) 
	cdef void _glTexGenfv "glTexGenfv"( GLenum coord, GLenum pname,  GLfloat *params ) 
	cdef void _glTexGeniv "glTexGeniv"( GLenum coord, GLenum pname,  GLint *params ) 
	cdef void _glGetTexGendv "glGetTexGendv"( GLenum coord, GLenum pname, GLdouble *params ) 
	cdef void _glGetTexGenfv "glGetTexGenfv"( GLenum coord, GLenum pname, GLfloat *params ) 
	cdef void _glGetTexGeniv "glGetTexGeniv"( GLenum coord, GLenum pname, GLint *params ) 
	cdef void _glTexEnvf "glTexEnvf"( GLenum target, GLenum pname, GLfloat param ) nogil
	cdef void _glTexEnvi "glTexEnvi"( GLenum target, GLenum pname, GLint param ) nogil
	cdef void _glTexEnvfv "glTexEnvfv"( GLenum target, GLenum pname,  GLfloat *params ) 
	cdef void _glTexEnviv "glTexEnviv"( GLenum target, GLenum pname,  GLint *params ) 
	cdef void _glGetTexEnvfv "glGetTexEnvfv"( GLenum target, GLenum pname, GLfloat *params ) 
	cdef void _glGetTexEnviv "glGetTexEnviv"( GLenum target, GLenum pname, GLint *params ) 
	cdef void _glTexParameterf "glTexParameterf"( GLenum target, GLenum pname, GLfloat param ) nogil
	cdef void _glTexParameteri "glTexParameteri"( GLenum target, GLenum pname, GLint param ) nogil
	cdef void _glTexParameterfv "glTexParameterfv"( GLenum target, GLenum pname,
                                           GLfloat *params ) 
	cdef void _glTexParameteriv "glTexParameteriv"( GLenum target, GLenum pname,
                                           GLint *params ) 
	cdef void _glGetTexParameterfv "glGetTexParameterfv"( GLenum target,
                                           GLenum pname, GLfloat *params) 
	cdef void _glGetTexParameteriv "glGetTexParameteriv"( GLenum target,
                                           GLenum pname, GLint *params ) 
	cdef void _glGetTexLevelParameterfv "glGetTexLevelParameterfv"( GLenum target, GLint level,
                                                GLenum pname, GLfloat *params ) 
	cdef void _glGetTexLevelParameteriv "glGetTexLevelParameteriv"( GLenum target, GLint level,
                                                GLenum pname, GLint *params ) 
	cdef void _glTexImage1D "glTexImage1D"( GLenum target, GLint level,
                                    GLint internalFormat,
                                    GLsizei width, GLint border,
                                    GLenum format, GLenum type,
                                     GLvoid *pixels ) 
	cdef void _glTexImage2D "glTexImage2D"( GLenum target, GLint level,
                                    GLint internalFormat,
                                    GLsizei width, GLsizei height,
                                    GLint border, GLenum format, GLenum type,
                                     GLvoid *pixels ) 
	cdef void _glGetTexImage "glGetTexImage"( GLenum target, GLint level,
                                     GLenum format, GLenum type,
                                     GLvoid *pixels ) 
	cdef void _glGenTextures "glGenTextures"( GLsizei n, GLuint *textures ) 
	cdef void _glDeleteTextures "glDeleteTextures"( GLsizei n,  GLuint *textures) 
	cdef void _glBindTexture "glBindTexture"( GLenum target, GLuint texture ) nogil
	cdef void _glPrioritizeTextures "glPrioritizeTextures"( GLsizei n,
                                             GLuint *textures,
                                             GLclampf *priorities ) 
	cdef GLboolean _glAreTexturesResident "glAreTexturesResident"( GLsizei n,
                                                   GLuint *textures,
                                                  GLboolean *residences ) 
	cdef GLboolean _glIsTexture "glIsTexture"( GLuint texture ) nogil
	cdef void _glTexSubImage1D "glTexSubImage1D"( GLenum target, GLint level,
                                       GLint xoffset,
                                       GLsizei width, GLenum format,
                                       GLenum type,  GLvoid *pixels ) 
	cdef void _glTexSubImage2D "glTexSubImage2D"( GLenum target, GLint level,
                                       GLint xoffset, GLint yoffset,
                                       GLsizei width, GLsizei height,
                                       GLenum format, GLenum type,
                                        GLvoid *pixels ) 
	cdef void _glCopyTexImage1D "glCopyTexImage1D"( GLenum target, GLint level,
                                        GLenum internalformat,
                                        GLint x, GLint y,
                                        GLsizei width, GLint border ) nogil
	cdef void _glCopyTexImage2D "glCopyTexImage2D"( GLenum target, GLint level,
                                        GLenum internalformat,
                                        GLint x, GLint y,
                                        GLsizei width, GLsizei height,
                                        GLint border ) nogil
	cdef void _glCopyTexSubImage1D "glCopyTexSubImage1D"( GLenum target, GLint level,
                                           GLint xoffset, GLint x, GLint y,
                                           GLsizei width ) nogil
	cdef void _glCopyTexSubImage2D "glCopyTexSubImage2D"( GLenum target, GLint level,
                                           GLint xoffset, GLint yoffset,
                                           GLint x, GLint y,
                                           GLsizei width, GLsizei height ) nogil
	cdef void _glMap1d "glMap1d"( GLenum target, GLdouble u1, GLdouble u2,
                               GLint stride,
                               GLint order,  GLdouble *points ) 
	cdef void _glMap1f "glMap1f"( GLenum target, GLfloat u1, GLfloat u2,
                               GLint stride,
                               GLint order,  GLfloat *points ) 
	cdef void _glMap2d "glMap2d"( GLenum target,
		     GLdouble u1, GLdouble u2, GLint ustride, GLint uorder,
		     GLdouble v1, GLdouble v2, GLint vstride, GLint vorder,
		      GLdouble *points ) 
	cdef void _glMap2f "glMap2f"( GLenum target,
		     GLfloat u1, GLfloat u2, GLint ustride, GLint uorder,
		     GLfloat v1, GLfloat v2, GLint vstride, GLint vorder,
		      GLfloat *points ) 
	cdef void _glGetMapdv "glGetMapdv"( GLenum target, GLenum query, GLdouble *v ) 
	cdef void _glGetMapfv "glGetMapfv"( GLenum target, GLenum query, GLfloat *v ) 
	cdef void _glGetMapiv "glGetMapiv"( GLenum target, GLenum query, GLint *v ) 
	cdef void _glEvalCoord1d "glEvalCoord1d"( GLdouble u ) nogil
	cdef void _glEvalCoord1f "glEvalCoord1f"( GLfloat u ) nogil
	cdef void _glEvalCoord1dv "glEvalCoord1dv"(  GLdouble *u ) 
	cdef void _glEvalCoord1fv "glEvalCoord1fv"(  GLfloat *u ) 
	cdef void _glEvalCoord2d "glEvalCoord2d"( GLdouble u, GLdouble v ) nogil
	cdef void _glEvalCoord2f "glEvalCoord2f"( GLfloat u, GLfloat v ) nogil
	cdef void _glEvalCoord2dv "glEvalCoord2dv"(  GLdouble *u ) 
	cdef void _glEvalCoord2fv "glEvalCoord2fv"(  GLfloat *u ) 
	cdef void _glMapGrid1d "glMapGrid1d"( GLint un, GLdouble u1, GLdouble u2 ) nogil
	cdef void _glMapGrid1f "glMapGrid1f"( GLint un, GLfloat u1, GLfloat u2 ) nogil
	cdef void _glMapGrid2d "glMapGrid2d"( GLint un, GLdouble u1, GLdouble u2,
                                   GLint vn, GLdouble v1, GLdouble v2 ) nogil
	cdef void _glMapGrid2f "glMapGrid2f"( GLint un, GLfloat u1, GLfloat u2,
                                   GLint vn, GLfloat v1, GLfloat v2 ) nogil
	cdef void _glEvalPoint1 "glEvalPoint1"( GLint i ) nogil
	cdef void _glEvalPoint2 "glEvalPoint2"( GLint i, GLint j ) nogil
	cdef void _glEvalMesh1 "glEvalMesh1"( GLenum mode, GLint i1, GLint i2 ) nogil
	cdef void _glEvalMesh2 "glEvalMesh2"( GLenum mode, GLint i1, GLint i2, GLint j1, GLint j2 ) nogil
	cdef void _glFogf "glFogf"( GLenum pname, GLfloat param ) nogil
	cdef void _glFogi "glFogi"( GLenum pname, GLint param ) nogil
	cdef void _glFogfv "glFogfv"( GLenum pname,  GLfloat *params ) 
	cdef void _glFogiv "glFogiv"( GLenum pname,  GLint *params ) 
	cdef void _glFeedbackBuffer "glFeedbackBuffer"( GLsizei size, GLenum type, GLfloat *buffer ) 
	cdef void _glPassThrough "glPassThrough"( GLfloat token ) nogil
	cdef void _glSelectBuffer "glSelectBuffer"( GLsizei size, GLuint *buffer ) 
	cdef void _glInitNames "glInitNames"() nogil
	cdef void _glLoadName "glLoadName"( GLuint name ) nogil
	cdef void _glPushName "glPushName"( GLuint name ) nogil
	cdef void _glPopName "glPopName"() nogil
	cdef void _glDrawRangeElements "glDrawRangeElements"( GLenum mode, GLuint start,
	GLuint end, GLsizei count, GLenum type,  GLvoid *indices ) 
	cdef void _glTexImage3D "glTexImage3D"( GLenum target, GLint level,
                                      GLint internalFormat,
                                      GLsizei width, GLsizei height,
                                      GLsizei depth, GLint border,
                                      GLenum format, GLenum type,
                                       GLvoid *pixels ) 
	cdef void _glTexSubImage3D "glTexSubImage3D"( GLenum target, GLint level,
                                         GLint xoffset, GLint yoffset,
                                         GLint zoffset, GLsizei width,
                                         GLsizei height, GLsizei depth,
                                         GLenum format,
                                         GLenum type,  GLvoid *pixels) 
	cdef void _glCopyTexSubImage3D "glCopyTexSubImage3D"( GLenum target, GLint level,
                                             GLint xoffset, GLint yoffset,
                                             GLint zoffset, GLint x,
                                             GLint y, GLsizei width,
                                             GLsizei height ) nogil
	cdef void _glColorTable "glColorTable"( GLenum target, GLenum internalformat,
                                    GLsizei width, GLenum format,
                                    GLenum type,  GLvoid *table ) 
	cdef void _glColorSubTable "glColorSubTable"( GLenum target,
                                       GLsizei start, GLsizei count,
                                       GLenum format, GLenum type,
                                        GLvoid *data ) 
	cdef void _glColorTableParameteriv "glColorTableParameteriv"(GLenum target, GLenum pname,
                                               GLint *params) 
	cdef void _glColorTableParameterfv "glColorTableParameterfv"(GLenum target, GLenum pname,
                                               GLfloat *params) 
	cdef void _glCopyColorSubTable "glCopyColorSubTable"( GLenum target, GLsizei start,
                                           GLint x, GLint y, GLsizei width ) nogil
	cdef void _glCopyColorTable "glCopyColorTable"( GLenum target, GLenum internalformat,
                                        GLint x, GLint y, GLsizei width ) nogil
	cdef void _glGetColorTable "glGetColorTable"( GLenum target, GLenum format,
                                       GLenum type, GLvoid *table ) 
	cdef void _glGetColorTableParameterfv "glGetColorTableParameterfv"( GLenum target, GLenum pname,
                                                  GLfloat *params ) 
	cdef void _glGetColorTableParameteriv "glGetColorTableParameteriv"( GLenum target, GLenum pname,
                                                  GLint *params ) 
	cdef void _glBlendEquation "glBlendEquation"( GLenum mode ) nogil
	cdef void _glBlendColor "glBlendColor"( GLclampf red, GLclampf green,
                                    GLclampf blue, GLclampf alpha ) nogil
	cdef void _glHistogram "glHistogram"( GLenum target, GLsizei width,
				   GLenum internalformat, GLboolean sink ) nogil
	cdef void _glResetHistogram "glResetHistogram"( GLenum target ) nogil
	cdef void _glGetHistogram "glGetHistogram"( GLenum target, GLboolean reset,
				      GLenum format, GLenum type,
				      GLvoid *values ) 
	cdef void _glGetHistogramParameterfv "glGetHistogramParameterfv"( GLenum target, GLenum pname,
						 GLfloat *params ) 
	cdef void _glGetHistogramParameteriv "glGetHistogramParameteriv"( GLenum target, GLenum pname,
						 GLint *params ) 
	cdef void _glMinmax "glMinmax"( GLenum target, GLenum internalformat,
				GLboolean sink ) nogil
	cdef void _glResetMinmax "glResetMinmax"( GLenum target ) nogil
	cdef void _glGetMinmax "glGetMinmax"( GLenum target, GLboolean reset,
                                   GLenum format, GLenum types,
                                   GLvoid *values ) 
	cdef void _glGetMinmaxParameterfv "glGetMinmaxParameterfv"( GLenum target, GLenum pname,
					      GLfloat *params ) 
	cdef void _glGetMinmaxParameteriv "glGetMinmaxParameteriv"( GLenum target, GLenum pname,
					      GLint *params ) 
	cdef void _glConvolutionFilter1D "glConvolutionFilter1D"( GLenum target,
	GLenum internalformat, GLsizei width, GLenum format, GLenum type,
	 GLvoid *image ) 
	cdef void _glConvolutionFilter2D "glConvolutionFilter2D"( GLenum target,
	GLenum internalformat, GLsizei width, GLsizei height, GLenum format,
	GLenum type,  GLvoid *image ) 
	cdef void _glConvolutionParameterf "glConvolutionParameterf"( GLenum target, GLenum pname,
	GLfloat params ) nogil
	cdef void _glConvolutionParameterfv "glConvolutionParameterfv"( GLenum target, GLenum pname,
	 GLfloat *params ) 
	cdef void _glConvolutionParameteri "glConvolutionParameteri"( GLenum target, GLenum pname,
	GLint params ) nogil
	cdef void _glConvolutionParameteriv "glConvolutionParameteriv"( GLenum target, GLenum pname,
	 GLint *params ) 
	cdef void _glCopyConvolutionFilter1D "glCopyConvolutionFilter1D"( GLenum target,
	GLenum internalformat, GLint x, GLint y, GLsizei width ) nogil
	cdef void _glCopyConvolutionFilter2D "glCopyConvolutionFilter2D"( GLenum target,
	GLenum internalformat, GLint x, GLint y, GLsizei width,
	GLsizei height) nogil
	cdef void _glGetConvolutionFilter "glGetConvolutionFilter"( GLenum target, GLenum format,
	GLenum type, GLvoid *image ) 
	cdef void _glGetConvolutionParameterfv "glGetConvolutionParameterfv"( GLenum target, GLenum pname,
	GLfloat *params ) 
	cdef void _glGetConvolutionParameteriv "glGetConvolutionParameteriv"( GLenum target, GLenum pname,
	GLint *params ) 
	cdef void _glSeparableFilter2D "glSeparableFilter2D"( GLenum target,
	GLenum internalformat, GLsizei width, GLsizei height, GLenum format,
	GLenum type,  GLvoid *row,  GLvoid *column ) 
	cdef void _glGetSeparableFilter "glGetSeparableFilter"( GLenum target, GLenum format,
	GLenum type, GLvoid *row, GLvoid *column, GLvoid *span ) 
	cdef void _glActiveTexture "glActiveTexture"( GLenum texture ) nogil
	cdef void _glClientActiveTexture "glClientActiveTexture"( GLenum texture ) nogil
	cdef void _glCompressedTexImage1D "glCompressedTexImage1D"( GLenum target, GLint level, GLenum internalformat, GLsizei width, GLint border, GLsizei imageSize,  GLvoid *data ) 
	cdef void _glCompressedTexImage2D "glCompressedTexImage2D"( GLenum target, GLint level, GLenum internalformat, GLsizei width, GLsizei height, GLint border, GLsizei imageSize,  GLvoid *data ) 
	cdef void _glCompressedTexImage3D "glCompressedTexImage3D"( GLenum target, GLint level, GLenum internalformat, GLsizei width, GLsizei height, GLsizei depth, GLint border, GLsizei imageSize,  GLvoid *data ) 
	cdef void _glCompressedTexSubImage1D "glCompressedTexSubImage1D"( GLenum target, GLint level, GLint xoffset, GLsizei width, GLenum format, GLsizei imageSize,  GLvoid *data ) 
	cdef void _glCompressedTexSubImage2D "glCompressedTexSubImage2D"( GLenum target, GLint level, GLint xoffset, GLint yoffset, GLsizei width, GLsizei height, GLenum format, GLsizei imageSize,  GLvoid *data ) 
	cdef void _glCompressedTexSubImage3D "glCompressedTexSubImage3D"( GLenum target, GLint level, GLint xoffset, GLint yoffset, GLint zoffset, GLsizei width, GLsizei height, GLsizei depth, GLenum format, GLsizei imageSize,  GLvoid *data ) 
	cdef void _glGetCompressedTexImage "glGetCompressedTexImage"( GLenum target, GLint lod, GLvoid *img ) 
	cdef void _glMultiTexCoord1d "glMultiTexCoord1d"( GLenum target, GLdouble s ) nogil
	cdef void _glMultiTexCoord1dv "glMultiTexCoord1dv"( GLenum target,  GLdouble *v ) 
	cdef void _glMultiTexCoord1f "glMultiTexCoord1f"( GLenum target, GLfloat s ) nogil
	cdef void _glMultiTexCoord1fv "glMultiTexCoord1fv"( GLenum target,  GLfloat *v ) 
	cdef void _glMultiTexCoord1i "glMultiTexCoord1i"( GLenum target, GLint s ) nogil
	cdef void _glMultiTexCoord1iv "glMultiTexCoord1iv"( GLenum target,  GLint *v ) 
	cdef void _glMultiTexCoord1s "glMultiTexCoord1s"( GLenum target, GLshort s ) nogil
	cdef void _glMultiTexCoord1sv "glMultiTexCoord1sv"( GLenum target,  GLshort *v ) 
	cdef void _glMultiTexCoord2d "glMultiTexCoord2d"( GLenum target, GLdouble s, GLdouble t ) nogil
	cdef void _glMultiTexCoord2dv "glMultiTexCoord2dv"( GLenum target,  GLdouble *v ) 
	cdef void _glMultiTexCoord2f "glMultiTexCoord2f"( GLenum target, GLfloat s, GLfloat t ) nogil
	cdef void _glMultiTexCoord2fv "glMultiTexCoord2fv"( GLenum target,  GLfloat *v ) 
	cdef void _glMultiTexCoord2i "glMultiTexCoord2i"( GLenum target, GLint s, GLint t ) nogil
	cdef void _glMultiTexCoord2iv "glMultiTexCoord2iv"( GLenum target,  GLint *v ) 
	cdef void _glMultiTexCoord2s "glMultiTexCoord2s"( GLenum target, GLshort s, GLshort t ) nogil
	cdef void _glMultiTexCoord2sv "glMultiTexCoord2sv"( GLenum target,  GLshort *v ) 
	cdef void _glMultiTexCoord3d "glMultiTexCoord3d"( GLenum target, GLdouble s, GLdouble t, GLdouble r ) nogil
	cdef void _glMultiTexCoord3dv "glMultiTexCoord3dv"( GLenum target,  GLdouble *v ) 
	cdef void _glMultiTexCoord3f "glMultiTexCoord3f"( GLenum target, GLfloat s, GLfloat t, GLfloat r ) nogil
	cdef void _glMultiTexCoord3fv "glMultiTexCoord3fv"( GLenum target,  GLfloat *v ) 
	cdef void _glMultiTexCoord3i "glMultiTexCoord3i"( GLenum target, GLint s, GLint t, GLint r ) nogil
	cdef void _glMultiTexCoord3iv "glMultiTexCoord3iv"( GLenum target,  GLint *v ) 
	cdef void _glMultiTexCoord3s "glMultiTexCoord3s"( GLenum target, GLshort s, GLshort t, GLshort r ) nogil
	cdef void _glMultiTexCoord3sv "glMultiTexCoord3sv"( GLenum target,  GLshort *v ) 
	cdef void _glMultiTexCoord4d "glMultiTexCoord4d"( GLenum target, GLdouble s, GLdouble t, GLdouble r, GLdouble q ) nogil
	cdef void _glMultiTexCoord4dv "glMultiTexCoord4dv"( GLenum target,  GLdouble *v ) 
	cdef void _glMultiTexCoord4f "glMultiTexCoord4f"( GLenum target, GLfloat s, GLfloat t, GLfloat r, GLfloat q ) nogil
	cdef void _glMultiTexCoord4fv "glMultiTexCoord4fv"( GLenum target,  GLfloat *v ) 
	cdef void _glMultiTexCoord4i "glMultiTexCoord4i"( GLenum target, GLint s, GLint t, GLint r, GLint q ) nogil
	cdef void _glMultiTexCoord4iv "glMultiTexCoord4iv"( GLenum target,  GLint *v ) 
	cdef void _glMultiTexCoord4s "glMultiTexCoord4s"( GLenum target, GLshort s, GLshort t, GLshort r, GLshort q ) nogil
	cdef void _glMultiTexCoord4sv "glMultiTexCoord4sv"( GLenum target,  GLshort *v ) 
	cdef void _glLoadTransposeMatrixd "glLoadTransposeMatrixd"(  GLdouble m[16] ) nogil
	cdef void _glLoadTransposeMatrixf "glLoadTransposeMatrixf"(  GLfloat m[16] ) nogil
	cdef void _glMultTransposeMatrixd "glMultTransposeMatrixd"(  GLdouble m[16] ) nogil
	cdef void _glMultTransposeMatrixf "glMultTransposeMatrixf"(  GLfloat m[16] ) nogil
	cdef void _glSampleCoverage "glSampleCoverage"( GLclampf value, GLboolean invert ) nogil
	cdef void _glActiveTextureARB "glActiveTextureARB"(GLenum texture) nogil
	cdef void _glClientActiveTextureARB "glClientActiveTextureARB"(GLenum texture) nogil
	cdef void _glMultiTexCoord1dARB "glMultiTexCoord1dARB"(GLenum target, GLdouble s) nogil
	cdef void _glMultiTexCoord1dvARB "glMultiTexCoord1dvARB"(GLenum target,  GLdouble *v) 
	cdef void _glMultiTexCoord1fARB "glMultiTexCoord1fARB"(GLenum target, GLfloat s) nogil
	cdef void _glMultiTexCoord1fvARB "glMultiTexCoord1fvARB"(GLenum target,  GLfloat *v) 
	cdef void _glMultiTexCoord1iARB "glMultiTexCoord1iARB"(GLenum target, GLint s) nogil
	cdef void _glMultiTexCoord1ivARB "glMultiTexCoord1ivARB"(GLenum target,  GLint *v) 
	cdef void _glMultiTexCoord1sARB "glMultiTexCoord1sARB"(GLenum target, GLshort s) nogil
	cdef void _glMultiTexCoord1svARB "glMultiTexCoord1svARB"(GLenum target,  GLshort *v) 
	cdef void _glMultiTexCoord2dARB "glMultiTexCoord2dARB"(GLenum target, GLdouble s, GLdouble t) nogil
	cdef void _glMultiTexCoord2dvARB "glMultiTexCoord2dvARB"(GLenum target,  GLdouble *v) 
	cdef void _glMultiTexCoord2fARB "glMultiTexCoord2fARB"(GLenum target, GLfloat s, GLfloat t) nogil
	cdef void _glMultiTexCoord2fvARB "glMultiTexCoord2fvARB"(GLenum target,  GLfloat *v) 
	cdef void _glMultiTexCoord2iARB "glMultiTexCoord2iARB"(GLenum target, GLint s, GLint t) nogil
	cdef void _glMultiTexCoord2ivARB "glMultiTexCoord2ivARB"(GLenum target,  GLint *v) 
	cdef void _glMultiTexCoord2sARB "glMultiTexCoord2sARB"(GLenum target, GLshort s, GLshort t) nogil
	cdef void _glMultiTexCoord2svARB "glMultiTexCoord2svARB"(GLenum target,  GLshort *v) 
	cdef void _glMultiTexCoord3dARB "glMultiTexCoord3dARB"(GLenum target, GLdouble s, GLdouble t, GLdouble r) nogil
	cdef void _glMultiTexCoord3dvARB "glMultiTexCoord3dvARB"(GLenum target,  GLdouble *v) 
	cdef void _glMultiTexCoord3fARB "glMultiTexCoord3fARB"(GLenum target, GLfloat s, GLfloat t, GLfloat r) nogil
	cdef void _glMultiTexCoord3fvARB "glMultiTexCoord3fvARB"(GLenum target,  GLfloat *v) 
	cdef void _glMultiTexCoord3iARB "glMultiTexCoord3iARB"(GLenum target, GLint s, GLint t, GLint r) nogil
	cdef void _glMultiTexCoord3ivARB "glMultiTexCoord3ivARB"(GLenum target,  GLint *v) 
	cdef void _glMultiTexCoord3sARB "glMultiTexCoord3sARB"(GLenum target, GLshort s, GLshort t, GLshort r) nogil
	cdef void _glMultiTexCoord3svARB "glMultiTexCoord3svARB"(GLenum target,  GLshort *v) 
	cdef void _glMultiTexCoord4dARB "glMultiTexCoord4dARB"(GLenum target, GLdouble s, GLdouble t, GLdouble r, GLdouble q) nogil
	cdef void _glMultiTexCoord4dvARB "glMultiTexCoord4dvARB"(GLenum target,  GLdouble *v) 
	cdef void _glMultiTexCoord4fARB "glMultiTexCoord4fARB"(GLenum target, GLfloat s, GLfloat t, GLfloat r, GLfloat q) nogil
	cdef void _glMultiTexCoord4fvARB "glMultiTexCoord4fvARB"(GLenum target,  GLfloat *v) 
	cdef void _glMultiTexCoord4iARB "glMultiTexCoord4iARB"(GLenum target, GLint s, GLint t, GLint r, GLint q) nogil
	cdef void _glMultiTexCoord4ivARB "glMultiTexCoord4ivARB"(GLenum target,  GLint *v) 
	cdef void _glMultiTexCoord4sARB "glMultiTexCoord4sARB"(GLenum target, GLshort s, GLshort t, GLshort r, GLshort q) nogil
	cdef void _glMultiTexCoord4svARB "glMultiTexCoord4svARB"(GLenum target,  GLshort *v) 
	cdef void _glGetProgramRegisterfvMESA "glGetProgramRegisterfvMESA"(GLenum target, GLsizei len,  GLubyte *name, GLfloat *v) 
	cdef void _glBlendEquationSeparateATI "glBlendEquationSeparateATI"( GLenum modeRGB, GLenum modeA ) nogil
	cdef void _glGenerateMipmap "glGenerateMipmap"(GLenum target) nogil
	cdef void _glGenerateMipmapEXT "glGenerateMipmapEXT"(GLenum target) nogil
	cdef void _glGenerateTextureMipmapEXT "glGenerateTextureMipmapEXT"(GLuint texture, GLenum target) nogil
	cdef void _glGenerateMultiTexMipmapEXT "glGenerateMultiTexMipmapEXT"(GLenum texunit, GLenum target) nogil
	cdef void _glGenBuffers "glGenBuffers"(GLsizei n, GLuint *buffers) 
	cdef void _glGenBuffersARB "glGenBuffersARB"(GLsizei n, GLuint *buffers) 
	cdef void _glBindBuffer "glBindBuffer"(GLenum target, GLuint buffer) nogil
	cdef void _glBindBufferRange "glBindBufferRange"(GLenum target, GLuint index, GLuint buffer, GLintptr offset, GLsizeiptr size) nogil
	cdef void _glBindBufferBase "glBindBufferBase"(GLenum target, GLuint index, GLuint buffer) nogil
	cdef void _glBindBufferARB "glBindBufferARB"(GLenum target, GLuint buffer) nogil
	cdef void _glBindBufferRangeNV "glBindBufferRangeNV"(GLenum target, GLuint index, GLuint buffer, GLintptr offset, GLsizeiptr size) nogil
	cdef void _glBindBufferOffsetNV "glBindBufferOffsetNV"(GLenum target, GLuint index, GLuint buffer, GLintptr offset) nogil
	cdef void _glBindBufferBaseNV "glBindBufferBaseNV"(GLenum target, GLuint index, GLuint buffer) nogil
	cdef void _glBindBufferRangeEXT "glBindBufferRangeEXT"(GLenum target, GLuint index, GLuint buffer, GLintptr offset, GLsizeiptr size) nogil
	cdef void _glBindBufferOffsetEXT "glBindBufferOffsetEXT"(GLenum target, GLuint index, GLuint buffer, GLintptr offset) nogil
	cdef void _glBindBufferBaseEXT "glBindBufferBaseEXT"(GLenum target, GLuint index, GLuint buffer) nogil
	cdef void _glBufferData "glBufferData"(GLenum target, GLsizeiptr size,  GLvoid *data, GLenum usage) 
	cdef void _glGenVertexArrays "glGenVertexArrays"(GLsizei n, GLuint *arrays) 
	cdef void _glBindVertexArray "glBindVertexArray"(GLuint array) nogil
	cdef void _glVertexAttrib4fv "glVertexAttrib4fv"(GLuint index,  GLfloat *v) 
	cdef void _glDeleteVertexArrays "glDeleteVertexArrays"(GLsizei n,  GLuint *arrays) 
	cdef void _glClearBufferfv "glClearBufferfv"(GLenum buffer, GLint drawbuffer,  GLfloat *value) 
	cdef void _glEnableVertexAttribArray "glEnableVertexAttribArray"(GLuint index) nogil
	cdef void _glVertexAttribPointer "glVertexAttribPointer"(GLuint index, GLint size, GLenum type, GLboolean normalized, GLsizei stride,  GLvoid *pointer) 
	cdef void _glValidateProgram "glValidateProgram"(GLuint program) nogil
	cdef void _glGetProgramInfoLog "glGetProgramInfoLog"(GLuint program, GLsizei bufSize, GLsizei *length, GLchar *infoLog) 
	cdef void _glDeleteProgram "glDeleteProgram"(GLuint program) nogil
	cdef GLuint _glCreateProgram "glCreateProgram"() nogil
	cdef GLuint _glCreateShader "glCreateShader"(GLenum type) nogil
	cdef void _glShaderSource "glShaderSource"(GLuint shader, GLsizei count,  GLchar* *string,  GLint *length) 
	cdef void _glCompileShader "glCompileShader"(GLuint shader) nogil
	cdef void _glAttachShader "glAttachShader"(GLuint program, GLuint shader) nogil
	cdef void _glLinkProgram "glLinkProgram"(GLuint program) nogil
	cdef void _glUseProgram "glUseProgram"(GLuint program) nogil
	cdef GLint _glGetUniformLocation "glGetUniformLocation"(GLuint program,  GLchar *name) 
	cdef void _glUniform1i "glUniform1i"(GLint location, GLint v0) nogil
	cdef void _glUniform4f "glUniform4f"(GLint location, GLfloat v0, GLfloat v1, GLfloat v2, GLfloat v3) nogil
	cdef void _glUniform1f "glUniform1f"(GLint location, GLfloat v0) nogil
	cdef GLint _glGetAttribLocation "glGetAttribLocation"(GLuint program,  GLchar *name) 
	cdef void _glBindFragDataLocation "glBindFragDataLocation"(GLuint program, GLuint color,  GLchar *name) 
	cdef void _glUniformMatrix4fv "glUniformMatrix4fv"(GLint location, GLsizei count, GLboolean transpose,  GLfloat *value) 



#!/usr/bin/env python
# encoding: utf-8

"""
Author(s): Matthew Loper

See LICENCE.txt for licensing and contact information.
"""

__all__ = ['ColoredRenderer', 'TexturedRenderer', 'DepthRenderer']

import numpy as np
from cvwrap import cv2
import time
import platform
import scipy.sparse as sp
from copy import deepcopy
import common
from common import draw_visibility_image, draw_barycentric_image, draw_colored_primitives, draw_texcoord_image
from topology import get_vertices_per_edge, get_faces_per_edge

if platform.system()=='Darwin':
    from contexts.ctx_mac import OsContext
else:
    from contexts.ctx_mesa import OsContext

from chumpy import *
from contexts._constants import *
from chumpy.utils import row, col


pixel_center_offset = 0.5


class BaseRenderer(Ch):
    terms = ['f', 'frustum','overdraw']
    dterms = ['camera', 'v']

    @depends_on('f') # not v: specifically, it depends only on the number of vertices, not on the values in v
    def primitives_per_edge(self):
        v=self.v.r.reshape((-1,3))
        f=self.f
        vpe = get_vertices_per_edge(v, f)
        fpe = get_faces_per_edge(v, f, vpe)
        return fpe, vpe

    @depends_on('f', 'frustum', 'camera', 'overdraw')
    def barycentric_image(self):
        self._call_on_changed()
        return draw_barycentric_image(self.glf, self.v.r, self.f, self.boundarybool_image if self.overdraw else None)

    @depends_on(terms+dterms)
    def boundaryid_image(self):
        self._call_on_changed()
        return draw_boundaryid_image(self.glb, self.v.r, self.f, self.vpe, self.fpe, self.camera)

    @depends_on('f', 'frustum', 'camera', 'overdraw')
    def visibility_image(self):
        self._call_on_changed()
        return draw_visibility_image(self.glb, self.v.r, self.f, self.boundarybool_image if self.overdraw else None)

    @depends_on(terms+dterms)
    def boundarybool_image(self):
        self._call_on_changed()
        boundaryid_image = self.boundaryid_image
        return np.asarray(boundaryid_image != 4294967295, np.uint32).reshape(boundaryid_image.shape)

    @property
    def shape(self):
        raise NotImplementedError('Should be implemented in inherited class.')

    @property
    def v(self):
        return self.camera.v

    @v.setter
    def v(self, newval):
        self.camera.v = newval

    @property
    def vpe(self):
        return self.primitives_per_edge[1]

    @property
    def fpe(self):
        return self.primitives_per_edge[0]


class DepthRenderer(BaseRenderer):
    terms = 'f', 'frustum', 'background_image','overdraw'
    dterms = 'camera', 'v'        
    

    @property
    def shape(self):
        return (self.frustum['height'], self.frustum['width'])

    def compute_r(self):
        tmp = self.camera.r
        return self.depth_image.reshape((self.frustum['height'], self.frustum['width']))
        
    def compute_dr_wrt(self, wrt):
        
        if wrt is not self.camera and wrt is not self.v:
            return None
        
        visibility = self.visibility_image
        visible = np.nonzero(visibility.ravel() != 4294967295)[0]
        barycentric = self.barycentric_image
        if wrt is self.camera:
            shape = visibility.shape
            depth = self.depth_image

            if self.overdraw:
                result1 = common.dImage_wrt_2dVerts_bnd(depth, visible, visibility, barycentric, self.frustum['width'], self.frustum['height'], self.v.r.size/3, self.f, self.boundaryid_image != 4294967295)
            else:
                result1 = common.dImage_wrt_2dVerts(depth, visible, visibility, barycentric, self.frustum['width'], self.frustum['height'], self.v.r.size/3, self.f)

            # result1 = common.dImage_wrt_2dVerts(depth, visible, visibility, barycentric, self.frustum['width'], self.frustum['height'], self.v.r.size/3, self.f)

            return result1

        elif wrt is self.v:

            IS = np.tile(col(visible), (1, 9)).ravel()
            JS = col(self.f[visibility.ravel()[visible]].ravel())
            JS = np.hstack((JS*3, JS*3+1, JS*3+2)).ravel()

            # FIXME: there should be a faster way to get the camera axis.
            # But it should be carefully tested with distortion present!
            pts = np.array([
                [self.camera.c.r[0], self.camera.c.r[1], 2],
                [self.camera.c.r[0], self.camera.c.r[1], 1]
            ])
            pts = self.camera.unproject_points(pts)
            cam_axis = pts[0,:] - pts[1,:]

            if True: # use barycentric coordinates (correct way)
                w = visibility.shape[1]
                pxs = np.asarray(visible % w, np.int32)
                pys = np.asarray(np.floor(np.floor(visible) / w), np.int32)
                bc0 = col(barycentric[pys, pxs, 0])
                bc1 = col(barycentric[pys, pxs, 1])
                bc2 = col(barycentric[pys, pxs, 2])
                bc = np.hstack((bc0,bc0,bc0,bc1,bc1,bc1,bc2,bc2,bc2)).ravel()
            else: # each vert contributes equally (an approximation)
                bc = 1. / 3.

            data = np.tile(row(cam_axis), (IS.size/3,1)).ravel() * bc
            result2 = sp.csc_matrix((data, (IS, JS)), shape=(self.frustum['height']*self.frustum['width'], self.v.r.size))
            return result2

    
    def on_changed(self, which):

        if 'frustum' in which:
            w = self.frustum['width']
            h = self.frustum['height']
            self.glf = OsContext(w, h, typ=GL_FLOAT)
            self.glf.Viewport(0, 0, w, h)
            self.glb = OsContext(w, h, typ=GL_UNSIGNED_BYTE)
            self.glb.Viewport(0, 0, w, h)
            
            
        if 'frustum' in which or 'camera' in which:
            setup_camera(self.glb, self.camera, self.frustum)
            setup_camera(self.glf, self.camera, self.frustum)

        if not hasattr(self, 'overdraw'):
            self.overdraw = True
            
        assert(self.v is self.camera.v)


    @depends_on(dterms+terms)
    def depth_image(self):
        self._call_on_changed()

        gl = self.glb
        gl.Clear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);

        gl.PolygonMode(GL_FRONT_AND_BACK, GL_FILL)
        draw_noncolored_verts(gl, self.camera.v.r, self.f)
        result = np.asarray(deepcopy(gl.getDepth()), np.float64)

        if self.overdraw:
            gl.PolygonMode(GL_FRONT_AND_BACK, GL_LINE)
            draw_noncolored_verts(gl, self.camera.v.r, self.f)
            overdraw = np.asarray(deepcopy(gl.getDepth()), np.float64)
            gl.PolygonMode(GL_FRONT_AND_BACK, GL_FILL)
            boundarybool_image = self.boundarybool_image
            result = overdraw*boundarybool_image + result*(1-boundarybool_image)

        if hasattr(self, 'background_image'):
            if False: # has problems at boundaries, not sure why yet
                bg_px = self.visibility_image == 4294967295
                fg_px = 1 - bg_px
                result = bg_px * self.background_image + fg_px * result
            else:
                tmp = np.concatenate((np.atleast_3d(result), np.atleast_3d(self.background_image)), axis=2)
                result = np.min(tmp, axis=2)

        return result

    def getDepthMesh(self, depth_image=None):
        self._call_on_changed() # make everything is up-to-date
        v = self.glb.getDepthCloud(depth_image)
        w = self.frustum['width']
        h = self.frustum['height']
        idxs = np.arange(w*h).reshape((h, w))
    
        # v0 is upperleft, v1 is upper right, v2 is lowerleft, v3 is lowerright
        v0 = col(idxs[:-1,:-1])
        v1 = col(idxs[:-1,1:])
        v2 = col(idxs[1:,:-1])
        v3 = col(idxs[1:,1:])

        f = np.hstack((v0, v1, v2, v1, v3, v2)).reshape((-1,3))        
        return v, f



class BoundaryRenderer(BaseRenderer):
    terms = 'f', 'frustum', 'num_channels'
    dterms = 'camera',

    @property
    def shape(self):
        return (self.frustum['height'], self.frustum['width'], self.num_channels)

    def compute_r(self):
        tmp = self.camera.r
        return self.color_image
    
    def compute_dr_wrt(self, wrt):
        if wrt is not self.camera:
            return None
        
        visibility = self.boundaryid_image
        shape = visibility.shape        
        
        visible = np.nonzero(visibility.ravel() != 4294967295)[0]
        num_visible = len(visible)

        barycentric = self.barycentric_image
    
        return common.dImage_wrt_2dVerts(self.color_image, visible, visibility, barycentric, self.frustum['width'], self.frustum['height'], self.v.r.size/3, self.vpe)

    def on_changed(self, which):
        if 'frustum' in which:
            w = self.frustum['width']
            h = self.frustum['height']
            self.glf = OsContext(w, h, typ=GL_FLOAT)
            self.glf.Viewport(0, 0, w, h)
            self.glb = OsContext(w, h, typ=GL_UNSIGNED_BYTE)
            self.glb.Viewport(0, 0, w, h)
            
        if 'frustum' in which or 'camera' in which:
            setup_camera(self.glb, self.camera, self.frustum)
            setup_camera(self.glf, self.camera, self.frustum)
            
        if not hasattr(self, 'overdraw'):
            self.overdraw = True
            
    @depends_on(terms+dterms)
    def color_image(self):
        self._call_on_changed()
        result = self.boundarybool_image.astype(np.float64)
        return np.dstack([result for i in range(self.num_channels)])


class ColoredRenderer(BaseRenderer):
    terms = 'f', 'frustum', 'background_image', 'overdraw', 'num_channels'
    dterms = 'vc', 'camera', 'bgcolor'        

    @property
    def shape(self):
        if not hasattr(self, 'num_channels'):
            self.num_channels = 3
        if self.num_channels > 1:
            return (self.frustum['height'], self.frustum['width'], self.num_channels)
        else:
            return (self.frustum['height'], self.frustum['width'])

    def compute_r(self):
        tmp = self.camera.r
        return self.color_image # .reshape((self.frustum['height'], self.frustum['width'], -1)).squeeze()
        
    def compute_dr_wrt(self, wrt):
        if wrt is not self.camera and wrt is not self.vc and wrt is not self.bgcolor:
            return None
        
        visibility = self.visibility_image

        shape = visibility.shape
        color = self.color_image
        
        visible = np.nonzero(visibility.ravel() != 4294967295)[0]
        num_visible = len(visible)

        barycentric = self.barycentric_image

        if wrt is self.camera:
            if self.overdraw:
                return common.dImage_wrt_2dVerts_bnd(color, visible, visibility, barycentric, self.frustum['width'], self.frustum['height'], self.v.r.size/3, self.f, self.boundaryid_image != 4294967295)
            else:
                return common.dImage_wrt_2dVerts(color, visible, visibility, barycentric, self.frustum['width'], self.frustum['height'], self.v.r.size/3, self.f)

        elif wrt is self.vc:
            return common.dr_wrt_vc(visible, visibility, self.f, barycentric, self.frustum, self.vc.size, num_channels=self.num_channels)

        elif wrt is self.bgcolor:
            return common.dr_wrt_bgcolor(visibility, self.frustum, num_channels=self.num_channels)


    def on_changed(self, which):
        if 'frustum' in which:
            w = self.frustum['width']
            h = self.frustum['height']
            self.glf = OsContext(w, h, typ=GL_FLOAT)
            self.glf.Viewport(0, 0, w, h)
            self.glb = OsContext(w, h, typ=GL_UNSIGNED_BYTE)
            self.glb.Viewport(0, 0, w, h)
            
        if 'frustum' in which or 'camera' in which:
            setup_camera(self.glb, self.camera, self.frustum)
            setup_camera(self.glf, self.camera, self.frustum)
            
        if not hasattr(self, 'num_channels'):
            self.num_channels = 3

        if not hasattr(self, 'bgcolor'):
            self.bgcolor = Ch(np.array([.5]*self.num_channels))
            which.add('bgcolor')

        if not hasattr(self, 'overdraw'):
            self.overdraw = True
            
        if 'bgcolor' in which or ('frustum' in which and hasattr(self, 'bgcolor')):
            self.glf.ClearColor(self.bgcolor.r[0], self.bgcolor.r[1%self.num_channels], self.bgcolor.r[2%self.num_channels], 1.)


    def flow_to(self, v_next, cam_next=None):
        return common.flow_to(self, v_next, cam_next)

    def filter_for_triangles(self, which_triangles):
        cim = self.color_image
        vim = self.visibility_image+1
        arr = np.zeros(len(self.f)+1)
        arr[which_triangles+1] = 1
        
        relevant_pixels = arr[vim.ravel()]
        cim2 = cim.copy() * np.atleast_3d(relevant_pixels.reshape(vim.shape))
        relevant_pixels = np.nonzero(arr[vim.ravel()])[0]
        xs = relevant_pixels % vim.shape[1]
        ys = relevant_pixels / vim.shape[1]
        return cim2[np.min(ys):np.max(ys), np.min(xs):np.max(xs), :]


    @depends_on('f', 'camera', 'vc')
    def boundarycolor_image(self):

        try:
            return self.draw_boundarycolor_image(with_vertex_colors=True)
        except:
            import pdb; pdb.set_trace()


    def draw_color_image(self, gl):
        self._call_on_changed()
        try:
            gl.Clear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);

            # use face colors if given
            # FIXME: this won't work for 2 channels
            draw_colored_verts(gl, self.v.r, self.f, self.vc.r)

            result = np.asarray(deepcopy(gl.getImage()[:,:,:self.num_channels].squeeze()), np.float64)

            if hasattr(self, 'background_image'):
                bg_px = np.tile(np.atleast_3d(self.visibility_image) == 4294967295, (1,1,self.num_channels)).squeeze()
                fg_px = 1 - bg_px
                result = bg_px * self.background_image + fg_px * result

            return result
        except:
            import pdb; pdb.set_trace()

    @depends_on(dterms+terms)
    def color_image(self):
        gl = self.glf
        gl.PolygonMode(GL_FRONT_AND_BACK, GL_FILL)
        no_overdraw = self.draw_color_image(gl)

        if not self.overdraw:
            return no_overdraw

        gl.PolygonMode(GL_FRONT_AND_BACK, GL_LINE)
        overdraw = self.draw_color_image(gl)
        gl.PolygonMode(GL_FRONT_AND_BACK, GL_FILL)

        boundarybool_image = self.boundarybool_image
        if self.num_channels > 1:
            boundarybool_image = np.atleast_3d(boundarybool_image)
        return np.asarray((overdraw*boundarybool_image + no_overdraw*(1-boundarybool_image)), order='C')

    @depends_on('f', 'frustum', 'camera')
    def boundary_images(self):
        self._call_on_changed()
        return draw_boundary_images(self.glb, self.v.r, self.f, self.vpe, self.fpe, self.camera)

    @depends_on(terms+dterms)    
    def boundarycolor_image(self): 
        self._call_on_changed()
        try:
            gl = self.glf
            colors = self.vc.r.reshape((-1,3))[self.vpe.ravel()]
            gl.Clear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);
            draw_colored_primitives(gl, self.v.r.reshape((-1,3)), self.vpe, colors)
            return np.asarray(deepcopy(gl.getImage()), np.float64)
        except:
            import pdb; pdb.set_trace()
            

        
class TexturedRenderer(ColoredRenderer):
    terms = 'f', 'frustum', 'vt', 'ft', 'background_image', 'overdraw'
    dterms = 'vc', 'camera', 'bgcolor', 'texture_image'

    def __del__(self):
        self.release_textures()

    @property
    def shape(self):
        return (self.frustum['height'], self.frustum['width'], 3)

    @property
    def num_channels(self):
        return 3
        
    def release_textures(self):
        if hasattr(self, 'textureID'):
            arr = np.asarray(np.array([self.textureID]), np.uint32)
            self.glf.DeleteTextures(arr)


    def compute_dr_wrt(self, wrt):
        result = super(TexturedRenderer, self).compute_dr_wrt(wrt)
        
        if wrt is self.vc:
            cim = self.draw_color_image(with_vertex_colors=False).ravel()
            cim = sp.spdiags(row(cim), [0], cim.size, cim.size)
            result = cim.dot(result)
        elif wrt is self.texture_image:
            IS = np.nonzero(self.visibility_image.ravel() != 4294967295)[0]
            JS = self.texcoord_image_quantized.ravel()[IS]

            clr_im = self.draw_color_image(with_vertex_colors=True, with_texture_on=False)

            if False:
                cv2.imshow('clr_im', clr_im)
                cv2.imshow('texmap', self.texture_image.r)
                cv2.waitKey(1)

            r = clr_im[:,:,0].ravel()[IS]
            g = clr_im[:,:,1].ravel()[IS]
            b = clr_im[:,:,2].ravel()[IS]
            data = np.concatenate((r,g,b))

            IS = np.concatenate((IS*3, IS*3+1, IS*3+2))
            JS = np.concatenate((JS*3, JS*3+1, JS*3+2))


            return sp.csc_matrix((data, (IS, JS)), shape=(self.r.size, wrt.r.size))

            
        return result

    def on_changed(self, which):
        super(TexturedRenderer, self).on_changed(which)

        # have to redo if frustum changes, b/c frustum triggers new context
        if 'texture_image' in which or 'frustum' in which:
            gl = self.glf
            texture_data = np.array(self.texture_image*255., dtype='uint8', order='C')
            tmp = np.zeros(2, dtype=np.uint32)
            
            self.release_textures()
            gl.GenTextures(1, tmp) # TODO: free after done
            self.textureID = tmp[0]

            gl.BindTexture(GL_TEXTURE_2D, self.textureID)

            gl.TexImage2Dub(GL_TEXTURE_2D, 0, GL_RGB, texture_data.shape[1], texture_data.shape[0], 0, GL_BGR, texture_data.ravel())
            #gl.Hint(GL_GENERATE_MIPMAP_HINT, GL_NICEST) # must be GL_FASTEST, GL_NICEST or GL_DONT_CARE
            gl.GenerateMipmap(GL_TEXTURE_2D)


    @depends_on('vt', 'ft')
    def mesh_tex_coords(self):
        ftidxs = self.ft.ravel()
        data = np.asarray(self.vt[ftidxs].astype(np.float32)[:,0:2], np.float32, order='C')
        data[:,1] = 1.0 - 1.0*data[:,1]
        return data
    
    # Depends on 'f' because vpe/fpe depend on f
    @depends_on('vt', 'ft', 'f')
    def wireframe_tex_coords(self):
        vvt = np.zeros((self.v.r.size/3,2), dtype=np.float32, order='C')
        vvt[self.f.flatten()] = self.mesh_tex_coords
        edata = np.zeros((self.vpe.size,2), dtype=np.float32, order='C')
        edata = vvt[self.vpe.ravel()]
        return edata
    
    def texture_mapping_on(self, gl, with_vertex_colors):
        gl.Enable(GL_TEXTURE_2D)
        gl.TexParameterf(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_LINEAR)
        gl.TexParameterf(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_LINEAR_MIPMAP_LINEAR)
        gl.TexEnvf(GL_TEXTURE_ENV, GL_TEXTURE_ENV_MODE, GL_MODULATE if with_vertex_colors else GL_REPLACE);
        gl.BindTexture(GL_TEXTURE_2D, self.textureID)
        gl.EnableClientState(GL_TEXTURE_COORD_ARRAY)

    def texture_mapping_off(self, gl):
        gl.Disable(GL_TEXTURE_2D)
        gl.DisableClientState(GL_TEXTURE_COORD_ARRAY)
    

    # TODO: can this not be inherited from base? turning off texture mapping in that instead?
    @depends_on(dterms+terms)
    def boundaryid_image(self):
        self._call_on_changed()
        self.texture_mapping_off(self.glb)
        result = draw_boundaryid_image(self.glb, self.v.r, self.f, self.vpe, self.fpe, self.camera)
        self.texture_mapping_on(self.glb, with_vertex_colors=True)
        return result


    @depends_on(terms+dterms)    
    def boundarycolor_image(self): 
        self._call_on_changed()
        try:
            gl = self.glf
            colors = self.vc.r.reshape((-1,3))[self.vpe.ravel()]
            gl.Clear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);
            self.texture_mapping_on(gl, with_vertex_colors=False if colors is None else True)
            gl.TexCoordPointerf(2,0, self.wireframe_tex_coords.ravel())
            draw_colored_primitives(self.glf, self.v.r.reshape((-1,3)), self.vpe, colors)
            self.texture_mapping_off(gl)
            return np.asarray(deepcopy(gl.getImage()), np.float64)
        except:
            import pdb; pdb.set_trace()

    def draw_color_image(self, with_vertex_colors=True, with_texture_on=True):
        self._call_on_changed()
        gl = self.glf
        gl.Clear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);

        if with_texture_on:
            self.texture_mapping_on(gl, with_vertex_colors)
            gl.TexCoordPointerf(2,0, self.mesh_tex_coords.ravel())
        else:
            self.texture_mapping_off(gl)

        colors = None
        if with_vertex_colors:
            colors = self.vc.r.reshape((-1,3))[self.f.ravel()]
        draw_colored_primitives(self.glf, self.v.r.reshape((-1,3)), self.f, colors)

        self.texture_mapping_off(gl)
        result =  np.asarray(deepcopy(gl.getImage()), np.float64)

        if hasattr(self, 'background_image'):
            bg_px = np.tile(np.atleast_3d(self.visibility_image) == 4294967295, (1,1,3))
            fg_px = 1 - bg_px
            result = bg_px * self.background_image + fg_px * result

        return result

    @depends_on('vt', 'ft', 'f', 'frustum', 'camera')
    def texcoord_image_quantized(self):
        texcoord_image = self.texcoord_image.copy()
        texcoord_image[:,:,0] *= self.texture_image.shape[1]-1
        texcoord_image[:,:,1] *= self.texture_image.shape[0]-1
        texcoord_image = np.round(texcoord_image)
        texcoord_image = texcoord_image[:,:,0] + texcoord_image[:,:,1]*self.texture_image.shape[1]
        return texcoord_image


    @depends_on('vt', 'ft', 'f', 'frustum', 'camera')
    def texcoord_image(self):
        return draw_texcoord_image(self.glf, self.v.r, self.f, self.vt, self.ft, self.boundarybool_image if self.overdraw else None)
        # gl = self.glf
        # self.texture_mapping_off(gl)
        # gl.Clear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);
        #
        # # want vtc: texture-coordinates per vertex (not per element in vc)
        # colors = self.vt[self.ft.ravel()]
        # #vtc = np.zeros((len(self.camera.r), 2))
        # #for idx, vidx in enumerate(self.f.ravel()):
        # #    tidx = self.ft.ravel()[idx]
        # #    vtc[vidx] = self.vt[tidx]
        #
        # colors = np.asarray(np.hstack((colors, col(colors[:,0]*0))), np.float64, order='C')
        # draw_colored_primitives(gl, self.camera.r.reshape((-1,3)), self.f, colors)
        # result = np.asarray(deepcopy(gl.getImage()), np.float64, order='C')[:,:,:2].copy()
        # result[:,:,1] = 1. - result[:,:,1]
        # return result




def draw_edge_visibility(gl, v, e, f, hidden_wireframe=True):
    """Assumes camera is set up correctly in gl context."""
    gl.Clear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);

    ec = np.arange(1, len(e)+1)
    ec = np.tile(col(ec), (1, 3))
    ec[:, 0] = ec[:, 0] & 255
    ec[:, 1] = (ec[:, 1] >> 8 ) & 255
    ec[:, 2] = (ec[:, 2] >> 16 ) & 255
    ec = np.asarray(ec, dtype=np.uint8)
    
    draw_colored_primitives(gl, v, e, ec)
    
    if hidden_wireframe:
        gl.Enable(GL_POLYGON_OFFSET_FILL)
        gl.PolygonOffset(10.0, 1.0)
        draw_colored_primitives(gl, v, f, fc=np.zeros(f.shape))
        gl.Disable(GL_POLYGON_OFFSET_FILL)
    
    raw = np.asarray(gl.getImage(), np.uint32)
    raw = raw[:,:,0] + raw[:,:,1]*256 + raw[:,:,2]*256*256 - 1
    return raw

def draw_boundary_images(glf, glb, v, f, vpe, fpe, camera):
    """Assumes camera is set up correctly, and that glf has any texmapping on necessary."""
    glf.Clear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);
    glb.Clear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);

    # Figure out which edges are on pairs of differently visible triangles
    from opendr.geometry import TriNormals
    tn = TriNormals(v, f).r.reshape((-1,3))
    campos = -cv2.Rodrigues(camera.rt.r)[0].T.dot(camera.t.r)
    rays_to_verts = v.reshape((-1,3)) - row(campos)
    rays_to_faces = rays_to_verts[f[:,0]] + rays_to_verts[f[:,1]] + rays_to_verts[f[:,2]]
    dps = np.sum(rays_to_faces * tn, axis=1)
    dps = dps[fpe[:,0]] * dps[fpe[:,1]]
    silhouette_edges = np.asarray(np.nonzero(dps<=0)[0], np.uint32)
    non_silhouette_edges = np.nonzero(dps>0)[0]
    lines_e = vpe[silhouette_edges]
    lines_v = v

    visibility = draw_edge_visibility(glb, lines_v, lines_e, f, hidden_wireframe=True)
    shape = visibility.shape
    visibility = visibility.ravel()
    visible = np.nonzero(visibility.ravel() != 4294967295)[0]
    visibility[visible] = silhouette_edges[visibility[visible]]
    result = visibility.reshape(shape)
    return result


def compute_vpe_boundary_idxs(v, f, camera, fpe):
    # Figure out which edges are on pairs of differently visible triangles

    from geometry import TriNormals
    tn = TriNormals(v, f).r.reshape((-1,3))

    #ray = cv2.Rodrigues(camera.rt.r)[0].T[:,2]
    campos = -cv2.Rodrigues(camera.rt.r)[0].T.dot(camera.t.r)
    rays_to_verts = v.reshape((-1,3)) - row(campos)
    rays_to_faces = rays_to_verts[f[:,0]] + rays_to_verts[f[:,1]] + rays_to_verts[f[:,2]]
    faces_invisible = np.sum(rays_to_faces * tn, axis=1)
    dps = faces_invisible[fpe[:,0]] * faces_invisible[fpe[:,1]]
    silhouette_edges = np.asarray(np.nonzero(dps<=0)[0], np.uint32)
    return silhouette_edges, faces_invisible < 0



def draw_boundaryid_image(gl, v, f, vpe, fpe, camera):


    if False:
        visibility = draw_edge_visibility(gl, v, vpe, f, hidden_wireframe=True)
        return visibility
        
    if True:
    #try:
        gl.Clear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);

        silhouette_edges, faces_facing_camera = compute_vpe_boundary_idxs(v, f, camera, fpe)
        lines_e = vpe[silhouette_edges]
        lines_v = v

        if len(lines_e)==0:
            return np.ones((gl.height, gl.width)).astype(np.int32) * 4294967295
        visibility = draw_edge_visibility(gl, lines_v, lines_e, f, hidden_wireframe=True)
        shape = visibility.shape
        visibility = visibility.ravel()
        visible = np.nonzero(visibility.ravel() != 4294967295)[0]
        visibility[visible] = silhouette_edges[visibility[visible]]
        result = visibility.reshape(shape)
        return result

        #return np.asarray(deepcopy(gl.getImage()), np.float64)
    #except:
    #    import pdb; pdb.set_trace()



def setup_camera(gl, camera, frustum):
    _setup_camera(gl,
                  camera.c.r[0], camera.c.r[1],
                  camera.f.r[0], camera.f.r[1],
                  frustum['width'], frustum['height'],
                  frustum['near'], frustum['far'],
                  camera.view_matrix,
                  camera.k.r)
    
    

vs_source = """
#version 120

uniform float k1, k2, k3, k4, k5, k6;
uniform float p1, p2;
uniform float cx, cy, fx, fy;

void main()
{
    vec4 p0 = gl_ModelViewMatrix * gl_Vertex;

    float xp = p0[0] / p0[2];
    float yp = -p0[1] / p0[2];

    float r2 = xp*xp + yp*yp;
    float r4 = r2 * r2;
    float r6 = r4 * r2;

    float m = (1.0 + k1*r2 + k2*r4 + k3*r6) / (1.0 + k4*r2 + k5*r4 + k6*r6);
    //p0[1] = -p0[1];
    p0[0] = xp * m + 2.*p1*xp*yp + p2*(r2+2*xp*xp);
    p0[1] = yp * m + p1*(r2+2*yp*yp) + 2.*p2*xp*yp;
    //p0[1] = -p0[1];
    p0[1] = -p0[1];

    gl_Position = gl_ProjectionMatrix * p0;
    //gl_Position = vec4(p0[0]*fx+cx, p0[1]*fy+cy, p0[2], p0[3]);
    //gl_Position[0] = p0[0]*fx+cx;
    //gl_Position[0] = p0[0];
    //gl_Position[0] = gl_Position[0] + 100;

    //----------------------------


    gl_FrontColor = gl_Color;
    gl_BackColor = gl_Color;

    //texture_coordinate = vec2(gl_MultiTexCoord0);
    gl_TexCoord[0] = gl_MultiTexCoord0;
}
"""

vs_source = """
#version 120

uniform float k1, k2, k3, k4, k5, k6;
uniform float p1, p2;

void main()
{
    vec4 p0 = gl_ModelViewMatrix * gl_Vertex;
    p0 = p0 / p0[3];

    float xp = -p0[0] / p0[2];
    float yp = p0[1] / p0[2];

    float r2 = xp*xp + yp*yp;
    float r4 = r2 * r2;
    float r6 = r4 * r2;

    float m = (1.0 + k1*r2 + k2*r4 + k3*r6) / (1.0 + k4*r2 + k5*r4 + k6*r6);

    float xpp = m*xp + 2.*p1*xp*yp + p2*(r2+2*xp*xp);
    float ypp = m*yp + p1*(r2+2*yp*yp) + 2.*p2*xp*yp;

    p0[0] = -xpp * p0[2];
    p0[1] = ypp * p0[2];
    gl_Position = gl_ProjectionMatrix * p0;

    //----------------------------

    gl_FrontColor = gl_Color;
    gl_BackColor = gl_Color;

    //texture_coordinate = vec2(gl_MultiTexCoord0);
    gl_TexCoord[0] = gl_MultiTexCoord0;
}
"""


def _setup_camera(gl, cx, cy, fx, fy, w, h, near, far, view_matrix, k):
    k = np.asarray(k)
    gl.MatrixMode(GL_PROJECTION)
    gl.LoadIdentity();
    
    f = 0.5 * (fx + fy)
    right  =  (w-(cx+pixel_center_offset)) * (near/f)
    left   =           -(cx+pixel_center_offset)  * (near/f)
    top    = -(h-(cy+pixel_center_offset)) * (near/f)
    bottom =            (cy+pixel_center_offset)  * (near/f)
    gl.Frustum(left, right, bottom, top, near, far)

    gl.MatrixMode(GL_MODELVIEW);
    gl.LoadIdentity(); # I
    gl.Rotatef(180, 1, 0, 0) # I * xR(pi)

    view_mtx = np.asarray(np.vstack((view_matrix, np.array([0, 0, 0, 1]))), np.float32, order='F')
    gl.MultMatrixf(view_mtx) # I * xR(pi) * V

    gl.Enable(GL_DEPTH_TEST)
    gl.PolygonMode(GL_FRONT_AND_BACK, GL_FILL)
    gl.Disable(GL_LIGHTING)
    gl.Disable(GL_CULL_FACE)
    gl.PixelStorei(GL_PACK_ALIGNMENT,1)
    gl.PixelStorei(GL_UNPACK_ALIGNMENT,1)

    if np.any(k):
        if not hasattr(gl, 'distortion_shader'):
            program = gl.CreateProgram()

            vs = gl.CreateShader(GL_VERTEX_SHADER)
            gl.ShaderSource(vs, 1, vs_source, len(vs_source))
            gl.AttachShader(program, vs)

            # fs = gl.CreateShader(GL_FRAGMENT_SHADER)
            # gl.ShaderSource(fs, 1, fs_source, len(fs_source))
            # gl.AttachShader(program, fs)

            gl.LinkProgram(program)
            gl.UseProgram(program)
            gl.distortion_shader = program

        gl.UseProgram(gl.distortion_shader)
        if len(k) != 8:
            tmp = k
            k = np.zeros(8)
            k[:len(tmp)] = tmp

        for idx, vname in enumerate(['k1', 'k2', 'p1', 'p2', 'k3', 'k4', 'k5', 'k6']):
            loc = gl.GetUniformLocation(gl.distortion_shader, vname)
            gl.Uniform1f(loc, k[idx])
    else:
        gl.UseProgram(0)


def draw_colored_verts(gl, v, f, vc):
    # TODO: copying is inefficient here
    if vc.shape[1] != 3:
        vc = np.vstack((vc[:,0], vc[:,1%vc.shape[1]], vc[:,2%vc.shape[1]])).T.copy()
    assert(vc.shape[1]==3)
    gl.EnableClientState(GL_VERTEX_ARRAY);
    gl.EnableClientState(GL_COLOR_ARRAY);
    gl.VertexPointer(np.ascontiguousarray(v).reshape((-1,3)));
    gl.ColorPointerd(np.ascontiguousarray(vc).reshape((-1,3)));
    gl.DrawElements(GL_TRIANGLES, np.asarray(f, np.uint32).ravel())

def draw_noncolored_verts(gl, v, f):
    gl.EnableClientState(GL_VERTEX_ARRAY);
    gl.DisableClientState(GL_COLOR_ARRAY);
    gl.VertexPointer(np.ascontiguousarray(v).reshape((-1,3)));
    gl.DrawElements(GL_TRIANGLES, np.asarray(f, np.uint32).ravel())

def main():
    pass

if __name__ == '__main__':
    main()




from cvwrap import cv2
import numpy as np
import chumpy as ch
from copy import deepcopy

def nothing(x):
    pass
    
    
def get_renderer():
    import chumpy as ch
    from opendr.everything import *

    # Load mesh
    m = load_mesh('/Users/matt/geist/OpenDR/test_dr/nasa_earth.obj')
    m.v += ch.array([0,0,3.])
    w, h = (320, 240)
    trans = ch.array([[0,0,0]])

    # Construct renderer
    rn = TexturedRenderer()
    rn.camera = ProjectPoints(v=m.v, rt=ch.zeros(3), t=ch.zeros(3), f=ch.array([w,w])/2., c=ch.array([w,h])/2., k=ch.zeros(5))
    rn.frustum = {'near': 1., 'far': 10., 'width': w, 'height': h}
    rn.set(v=trans+m.v, f=m.f, texture_image=m.texture_image[:,:,::-1], ft=m.ft, vt=m.vt, bgcolor=ch.zeros(3))
    rn.vc = SphericalHarmonics(vn=VertNormals(v=rn.v, f=rn.f), components=ch.array([4.,0.,0.,0.]), light_color=ch.ones(3))

    return rn
    
    
def main():
    # Create a black image, a window
    img = np.zeros((300,512,3), np.uint8)
    cv2.namedWindow('image')
    cv2.namedWindow('derivatives')

    rn = get_renderer()

    tracked = {
        'sph0': rn.vc.components[0],
        'sph1': rn.vc.components[1],
        'sph2': rn.vc.components[2],
        'sph3': rn.vc.components[3],
        'k0': rn.camera.k[0],
        'k1': rn.camera.k[1],
        'k2': rn.camera.k[2]
    }

    cnst = 1000
    for k in sorted(tracked.keys()):
        v = tracked[k]
        cv2.createTrackbar(k, 'image', 0,cnst, nothing)

    old_tracked = tracked
    cv2.setTrackbarPos('sph0', 'image', 800)
    while(1):
        cv2.imshow('image',rn.r)
        k = cv2.waitKey(1) & 0xFF
        if k == 27:
            break
    
        for k, v in tracked.items():
            v[:] = np.array(cv2.getTrackbarPos(k, 'image')).astype(np.float32)*4/cnst
            if tracked[k].r[0] != old_tracked[k].r[0]:
                drim = rn.dr_wrt(v).reshape(rn.shape)
                mn = np.mean(drim)
                drim /= np.max(np.abs(drim.ravel()))*2.
                drim += .5
                # drim = drim - np.min(drim)
                # drim = drim / np.max(drim)
                cv2.imshow('derivatives', drim)
               
        cv2.waitKey(1)
        old_tracked = deepcopy(tracked)
    # while True: 
    #     for k_change in sorted(tracked.keys()):
    #         if k_change == 'sph0':
    #             continue
    #         for t in np.arange(0, np.pi, .05):
    #             cv2.setTrackbarPos(k_change, 'image', int(np.sin(t)*1000))
    #             cv2.imshow('image',rn.r)
    #             k = cv2.waitKey(1) & 0xFF
    #             if k == 27:
    #                 break
    # 
    #             for k, v in tracked.items():
    #                 v[:] = np.array(cv2.getTrackbarPos(k, 'image')).astype(np.float32)*4/cnst
    #                 if tracked[k].r[0] != old_tracked[k].r[0]:
    #                     drim = rn.dr_wrt(v).reshape(rn.shape)
    #                     mn = np.mean(drim)
    #                     drim /= np.max(np.abs(drim.ravel()))*2.
    #                     drim += .5
    #                     # drim = drim - np.min(drim)
    #                     # drim = drim / np.max(drim)
    #                     cv2.imshow('derivatives', drim)
    #         
    #         
    #             print rn.vc.components
    # 
    #             cv2.waitKey(1)
    #             old_tracked = deepcopy(tracked)

    cv2.destroyAllWindows()


if __name__ == '__main__':
    main()



#!/usr/bin/env python

"""
Author(s): Matthew Loper

See LICENCE.txt for licensing and contact information.
"""

__all__ = ['dummy']

class Dummy(object):
    def __getattr__(self, name):
        def wrapper(*args, **kwargs):
            pass
        return wrapper

dummy = Dummy()


class Minimal(object):
    def __init__(self, **kwargs):
        self.__dict__ = kwargs


from version import version as __version__

def test():
    from os.path import split
    import unittest
    test_loader = unittest.TestLoader()
    test_loader = test_loader.discover(split(__file__)[0])
    test_runner = unittest.TextTestRunner()
    test_runner.run( test_loader )

demos = {}

demos['texture'] = """
# Create renderer
import chumpy as ch
from opendr.renderer import TexturedRenderer
rn = TexturedRenderer()

# Assign attributes to renderer
from opendr.util_tests import get_earthmesh
m = get_earthmesh(trans=ch.array([0,0,4]), rotation=ch.zeros(3))
w, h = (320, 240)
from opendr.camera import ProjectPoints
rn.camera = ProjectPoints(v=m.v, rt=ch.zeros(3), t=ch.zeros(3), f=ch.array([w,w])/2., c=ch.array([w,h])/2., k=ch.zeros(5))
rn.frustum = {'near': 1., 'far': 10., 'width': w, 'height': h}
rn.set(v=m.v, f=m.f, vc=m.vc, texture_image=m.texture_image, ft=m.ft, vt=m.vt)

# Show it
import matplotlib.pyplot as plt
plt.ion()
plt.imshow(rn.r)
plt.show()

dr = rn.dr_wrt(rn.v) # or rn.vc, or rn.camera.rt, rn.camera.t, rn.camera.f, rn.camera.c, etc
"""

demos['moments'] = """
from opendr.util_tests import get_earthmesh
from opendr.simple import * 
import numpy as np

w, h = 320, 240

m = get_earthmesh(trans=ch.array([0,0,4]), rotation=ch.zeros(3))

# Create V, A, U, f: geometry, brightness, camera, renderer
V = ch.array(m.v)
A = SphericalHarmonics(vn=VertNormals(v=V, f=m.f), 
                       components=[3.,1.,0.,0.,0.,0.,0.,0.,0.], 
                       light_color=ch.ones(3))
U = ProjectPoints(v=V, f=[300,300.], c=[w/2.,h/2.], k=ch.zeros(5),
                  t=ch.zeros(3), rt=ch.zeros(3))
rn = TexturedRenderer(vc=A, camera=U, f=m.f, bgcolor=[0.,0.,0.],
                     texture_image=m.texture_image, vt=m.vt, ft=m.ft,
                     frustum={'width':w, 'height':h, 'near':1,'far':20})

i, j = ch.array([2.]), ch.array([1.])
xs, ys = ch.meshgrid(range(rn.shape[1]), range(rn.shape[0]))
ysp = ys ** j
xsp = xs ** i
rn_bw = ch.sum(rn, axis=2)
moment = ch.sum((rn_bw * ysp * xsp).ravel())

# Print our numerical result
print moment 

# Note that opencv produces the same result for 'm21',
# and that other moments can be created by changing "i" and "j" above
import cv2
print cv2.moments(rn_bw.r)['m21']

# Derivatives wrt vertices and lighting
print moment.dr_wrt(V)
print moment.dr_wrt(A.components)
"""

demos['per_face_normals'] = """
# Create renderer
import chumpy as ch
import numpy as np
from opendr.renderer import ColoredRenderer
from opendr.lighting import LambertianPointLight
rn = ColoredRenderer()

# Assign attributes to renderer
from opendr.util_tests import get_earthmesh
m = get_earthmesh(trans=ch.array([0,0,4]), rotation=ch.zeros(3))
w, h = (320, 240)

# THESE ARE THE 3 CRITICAL LINES 
m.v = m.v[m.f.ravel()]
m.vc = m.vc[m.f.ravel()]
m.f = np.arange(m.f.size).reshape((-1,3))

from opendr.camera import ProjectPoints
rn.camera = ProjectPoints(v=m.v, rt=ch.zeros(3), t=ch.zeros(3), f=ch.array([w,w])/2., c=ch.array([w,h])/2., k=ch.zeros(5))
rn.frustum = {'near': 1., 'far': 10., 'width': w, 'height': h}
rn.set(v=m.v, f=m.f, bgcolor=ch.zeros(3))

# Construct point light source
rn.vc = LambertianPointLight(
    f=m.f,
    v=rn.v,
    num_verts=len(m.v),
    light_pos=ch.array([-1000,-1000,-1000]),
    vc=m.vc,
    light_color=ch.array([1., 1., 1.]))

# Show it
import matplotlib.pyplot as plt
plt.ion()
plt.imshow(rn.r)
plt.show()

dr = rn.dr_wrt(rn.v) # or rn.vc, or rn.camera.rt, rn.camera.t, rn.camera.f, rn.camera.c, etc
"""

demos['silhouette'] = """
# Create renderer
import chumpy as ch
from opendr.renderer import ColoredRenderer
rn = ColoredRenderer()

# Assign attributes to renderer
from opendr.util_tests import get_earthmesh
m = get_earthmesh(trans=ch.array([0,0,4]), rotation=ch.zeros(3))
w, h = (320, 240)
from opendr.camera import ProjectPoints
rn.camera = ProjectPoints(v=m.v, rt=ch.zeros(3), t=ch.zeros(3), f=ch.array([w,w])/2., c=ch.array([w,h])/2., k=ch.zeros(5))
rn.frustum = {'near': 1., 'far': 10., 'width': w, 'height': h}
rn.set(v=m.v, f=m.f, vc=m.vc*0+1, bgcolor=ch.zeros(3))

# Show it
import matplotlib.pyplot as plt
plt.ion()
plt.imshow(rn.r)
plt.show()

dr = rn.dr_wrt(rn.v) # or rn.vc, or rn.camera.rt, rn.camera.t, rn.camera.f, rn.camera.c, etc
"""

demos['boundary'] = """
# Create renderer
import chumpy as ch
from opendr.renderer import BoundaryRenderer
rn = BoundaryRenderer()

# Assign attributes to renderer
from opendr.util_tests import get_earthmesh
m = get_earthmesh(trans=ch.array([0,0,4]), rotation=ch.zeros(3))
w, h = (320, 240)
from opendr.camera import ProjectPoints
rn.camera = ProjectPoints(v=m.v, rt=ch.zeros(3), t=ch.zeros(3), f=ch.array([w,w])/2., c=ch.array([w,h])/2., k=ch.zeros(5))
rn.frustum = {'near': 1., 'far': 10., 'width': w, 'height': h}
rn.set(v=m.v, f=m.f, vc=m.vc*0+1, bgcolor=ch.zeros(3), num_channels=3)

# Show it
import matplotlib.pyplot as plt
plt.ion()
plt.imshow(rn.r)
plt.show()

dr = rn.dr_wrt(rn.v) # or rn.vc, or rn.camera.rt, rn.camera.t, rn.camera.f, rn.camera.c, etc
"""

demos['point_light'] = """
# Create renderer
import chumpy as ch
from opendr.renderer import ColoredRenderer
from opendr.lighting import LambertianPointLight
rn = ColoredRenderer()

# Assign attributes to renderer
from opendr.util_tests import get_earthmesh
m = get_earthmesh(trans=ch.array([0,0,4]), rotation=ch.zeros(3))
w, h = (320, 240)

from opendr.camera import ProjectPoints
rn.camera = ProjectPoints(v=m.v, rt=ch.zeros(3), t=ch.zeros(3), f=ch.array([w,w])/2., c=ch.array([w,h])/2., k=ch.zeros(5))
rn.frustum = {'near': 1., 'far': 10., 'width': w, 'height': h}
rn.set(v=m.v, f=m.f, bgcolor=ch.zeros(3))

# Construct point light source
rn.vc = LambertianPointLight(
    f=m.f,
    v=rn.v,
    num_verts=len(m.v),
    light_pos=ch.array([-1000,-1000,-1000]),
    vc=m.vc,
    light_color=ch.array([1., 1., 1.]))

# Show it
import matplotlib.pyplot as plt
plt.ion()
plt.imshow(rn.r)
plt.show()

dr = rn.dr_wrt(rn.v) # or rn.vc, or rn.camera.rt, rn.camera.t, rn.camera.f, rn.camera.c, etc
"""

demos['spherical_harmonics'] = """
# Create renderer
import chumpy as ch
from opendr.renderer import ColoredRenderer
from opendr.lighting import SphericalHarmonics
from opendr.geometry import VertNormals

rn = ColoredRenderer()

# Assign attributes to renderer
from opendr.util_tests import get_earthmesh
m = get_earthmesh(trans=ch.array([0,0,4]), rotation=ch.zeros(3))
w, h = (320, 240)
from opendr.camera import ProjectPoints
rn.camera = ProjectPoints(v=m.v, rt=ch.zeros(3), t=ch.zeros(3), f=ch.array([w,w])/2., c=ch.array([w,h])/2., k=ch.zeros(5))
rn.frustum = {'near': 1., 'far': 10., 'width': w, 'height': h}
rn.set(v=m.v, f=m.f, bgcolor=ch.zeros(3))

vn = VertNormals(v=rn.v, f=rn.f)
sh_red = SphericalHarmonics(vn=vn, light_color=ch.array([1,0,0]), components=ch.random.randn(9))
sh_green = SphericalHarmonics(vn=vn, light_color=ch.array([0,1,0]), components=ch.random.randn(9))
sh_blue = SphericalHarmonics(vn=vn, light_color=ch.array([0,0,1]), components=ch.random.randn(9))
rn.vc = sh_red + sh_green + sh_blue

# Show it
import matplotlib.pyplot as plt
plt.ion()
plt.imshow(rn.r)
plt.show()

dr = rn.dr_wrt(rn.v) # or rn.vc, or rn.camera.rt, rn.camera.t, rn.camera.f, rn.camera.c, etc
"""

demos['optimization'] = """
from opendr.simple import *      
import numpy as np 
import matplotlib.pyplot as plt
w, h = 320, 240
        
try:
    m = load_mesh('earth.obj')
except:                                                             
    from opendr.util_tests import get_earthmesh
    m = get_earthmesh(trans=ch.array([0,0,0]), rotation=ch.zeros(3))                                             
                                                                                
# Create V, A, U, f: geometry, brightness, camera, renderer                     
V = ch.array(m.v)                                                               
A = SphericalHarmonics(vn=VertNormals(v=V, f=m.f),                              
                       components=[3.,2.,0.,0.,0.,0.,0.,0.,0.],
                       light_color=ch.ones(3))                                  
U = ProjectPoints(v=V, f=[w,w], c=[w/2.,h/2.], k=ch.zeros(5),              
                  t=ch.zeros(3), rt=ch.zeros(3))                                
f = TexturedRenderer(vc=A, camera=U, f=m.f, bgcolor=[0.,0.,0.],                 
                     texture_image=m.texture_image, vt=m.vt, ft=m.ft,           
                     frustum={'width':w, 'height':h, 'near':1,'far':20})     


# Parameterize the vertices                                                     
translation, rotation = ch.array([0,0,8]), ch.zeros(3)                          
f.v = translation + V.dot(Rodrigues(rotation))                                  
           
observed = f.r 
np.random.seed(1)
translation[:] = translation.r + np.random.rand(3)
rotation[:] = rotation.r + np.random.rand(3) *.2
A.components[1:] = 0

# Create the energy
E_raw = f - observed
E_pyr = gaussian_pyramid(E_raw, n_levels=6, normalization='size')

def cb(_):
    import cv2
    global E_raw
    cv2.imshow('Absolute difference', np.abs(E_raw.r))
    cv2.waitKey(1)                  

print 'OPTIMIZING TRANSLATION, ROTATION, AND LIGHT PARMS'    
free_variables=[translation, rotation, A.components]
ch.minimize({'pyr': E_pyr}, x0=free_variables, callback=cb) 
ch.minimize({'raw': E_raw}, x0=free_variables, callback=cb) 

"""

demos['optimization_cpl'] = """
from opendr.simple import *      
import numpy as np 
import matplotlib.pyplot as plt
w, h = 320, 240
        
try:
    m = load_mesh('earth.obj')
except:                                                             
    from opendr.util_tests import get_earthmesh
    m = get_earthmesh(trans=ch.array([0,0,0]), rotation=ch.zeros(3))                                             
                                                                                
# Create V, A, U, f: geometry, brightness, camera, renderer                     
V = ch.array(m.v)                                                               
A = SphericalHarmonics(vn=VertNormals(v=V, f=m.f),                              
                       components=[3.,2.,0.,0.,0.,0.,0.,0.,0.],
                       light_color=ch.ones(3))                                  
U = ProjectPoints(v=V, f=[w,w], c=[w/2.,h/2.], k=ch.zeros(5),              
                  t=ch.zeros(3), rt=ch.zeros(3))                                
f = TexturedRenderer(vc=A, camera=U, f=m.f, bgcolor=[0.,0.,0.],                 
                     texture_image=m.texture_image, vt=m.vt, ft=m.ft,           
                     frustum={'width':w, 'height':h, 'near':1,'far':20})     


# Parameterize the vertices                                                     
translation, rotation = ch.array([0,0,8]), ch.zeros(3)   
model_v = translation + ch.array(V.r).dot(Rodrigues(rotation))   

# Simulate an observed image                               
V[:] = model_v.r
observed = f.r 

np.random.seed(1)
translation[:] = translation.r + np.random.rand(3)
rotation[:] = rotation.r + np.random.rand(3) *.2
V[:] = model_v.r
A.components[1:] = 0

# Create the energy
E_raw = f - observed
E_pyr = gaussian_pyramid(E_raw, n_levels=6, normalization='size')

def cb(_):
    import cv2
    global E_raw
    cv2.imshow('Absolute difference', np.abs(E_raw.r))
    cv2.waitKey(1)                  

print 'OPTIMIZING TRANSLATION, ROTATION, AND LIGHT PARMS'    
free_variables=[translation, rotation, A.components, V]
ch.minimize({'pyr': E_pyr, 'cpl': (V - model_v)*1e-4}, x0=free_variables, callback=cb) 
ch.minimize({'raw': E_raw, 'cpl': V - model_v}, x0=free_variables, callback=cb) 
"""

def demo(which=None):
    import re
    if which not in demos:
        print 'Please indicate which demo you want, as follows:'
        for key in demos:
            print "\tdemo('%s')" % (key,)
        return

    print '- - - - - - - - - - - <CODE> - - - - - - - - - - - -'
    print re.sub('global.*\n','',demos[which])
    print '- - - - - - - - - - - </CODE> - - - - - - - - - - - -\n'
    exec('global np\n' + demos[which], globals(), locals())



"""
Author(s): Matthew Loper

See LICENCE.txt for licensing and contact information.
"""
__all__ = ['mstack', 'wget']

def mstack(vs, fs):
    import chumpy as ch
    import numpy as np
    lengths = [v.shape[0] for v in vs]
    f = np.vstack([fs[i]+np.sum(lengths[:i]).astype(np.uint32) for i in range(len(fs))])
    v = ch.vstack(vs)

    return v, f


def wget(url, dest_fname=None):
    import urllib2
    from os.path import split, join

    curdir = split(__file__)[0]
    if dest_fname is None:
        dest_fname = join(curdir, split(url)[1])

    try:
        contents = urllib2.urlopen(url).read()
    except:
        raise Exception('Unable to get url: %s' % (url,))
    open(dest_fname, 'w').write(contents)



"""
Author(s): Matthew Loper

See LICENCE.txt for licensing and contact information.
"""

__author__ = 'matt'

__all__ = ['load_mesh', 'load_image']

from os.path import split, splitext, join, exists, normpath
from cvwrap import cv2
import numpy as np
from dummy import Minimal


def load_image(filename):
    return (cv2.imread(filename)[:,:,::-1]/255.).copy()

def load_mesh(filename):

    extension = splitext(filename)[1]
    if  extension == '.ply':
        return read_ply(filename)
    elif extension == '.obj':
        return read_obj(filename)
    else:
        raise Exception('Unsupported file extension for %s' % (filename,))

def _update_mtl(mtl, filename):

    lines = [l.strip() for l in open(filename).read().split('\n')]

    curkey = ''
    for line in lines:
        spl = line.split()

        if len(spl) < 2:
            continue
        key = spl[0]
        values = spl[1:]

        if key == 'newmtl':
            curkey = values[0]
            mtl[curkey] = {'filename': filename}
        elif curkey:
            mtl[curkey][key] = values


def read_obj(filename):

    obj_directory = split(filename)[0]
    lines = open(filename).read().split('\n')

    d = {'v': [], 'vn': [], 'f': [], 'vt': [], 'ft': []}

    mtls = {}
    for line in lines:
        line = line.split()
        if len(line) < 2:
            continue

        key = line[0]
        values = line[1:]

        if key == 'v':
            d['v'].append([np.array([float(v) for v in values[:3]])])
        elif key == 'f':
            spl = [l.split('/') for l in values]
            d['f'].append([np.array([int(l[0])-1 for l in spl[:3]], dtype=np.uint32)])
            if len(spl) > 1 and spl[1] and 'ft' in d:
                d['ft'].append([np.array([int(l[1])-1 for l in spl[:3]])])

            # TOO: redirect to actual vert normals?
            #if len(line[0]) > 2 and line[0][2]:
            #    d['fn'].append([np.concatenate([l[2] for l in spl[:3]])])
        elif key == 'vn':
            d['vn'].append([np.array([float(v) for v in values])])
        elif key == 'vt':
            d['vt'].append([np.array([float(v) for v in values])])
        elif key == 'mtllib':
            fname = join(obj_directory, values[0])
            if not exists(fname):
                fname = values[0]
            if not exists(fname):
                raise Exception("Can't find path %s" % (values[0]))
            _update_mtl(mtls, fname)
        elif key == 'usemtl':
            cur_mtl = mtls[values[0]]

            if 'map_Kd' in cur_mtl:
                src_fname = cur_mtl['map_Kd'][0]
                dst_fname = join(split(cur_mtl['filename'])[0], src_fname)
                if not exists(dst_fname):
                    dst_fname = join(obj_directory, src_fname)
                if not exists(dst_fname):
                    dst_fname = src_fname
                if not exists(dst_fname):
                    raise Exception("Unable to find referenced texture map %s" % (src_fname,))
                else:
                    d['texture_filepath'] = normpath(dst_fname)
                    im = cv2.imread(dst_fname)
                    sz = np.sqrt(np.prod(im.shape[:2]))
                    sz = int(np.round(2 ** np.ceil(np.log(sz) / np.log(2))))
                    d['texture_image'] = cv2.resize(im, (sz, sz)).astype(np.float64)/255.

    for k, v in d.items():
        if k in ['v','vn','f','vt','ft']:
            if v:
                d[k] = np.vstack(v)
            else:
                del d[k]
        else:
            d[k] = v

    result = Minimal(**d)

    return result




def read_ply(filename):

    import numpy as np
    str2dtype = {
        'char':   np.int8,
        'uchar':  np.uint8,
        'short':  np.int16,
        'ushort': np.uint16,
        'int':    np.int32,
        'uint':   np.uint32,
        'float':  np.float32,
        'double': np.float64
    }

    data = open(filename, 'rb').read()

    if data[:3] != 'ply':
        raise Exception('Need ply header.')

    if data[3:5] == '\r\n':
        newline = '\r\n'
    elif data[3] == '\r':
        newline = '\r'
    elif data[3] == '\n':
        newline = '\n'
    else:
        raise Exception('Bad newline after ply header.')


    pos = data.find('end_header' + newline)
    header = data[:pos].split(newline)
    body = data[pos + len('end_header' + newline):]

    elements = []
    for line in header:
        tokens = [ln for ln in line.split(' ') if len(ln) > 0]
        if len(tokens) <= 0:
            continue
        if tokens[0] == 'format':
            format = tokens[1]

        elif tokens[0] == 'element':
            elements.append({'name': tokens[1], 'len': int(tokens[2]), 'properties': []})

        elif tokens[0] == 'property':
            prop = {'name': tokens[-1], 'dtype': str2dtype[tokens[-2]]}
            if tokens[1] == 'list':
                prop['list_dtype'] = str2dtype[tokens[-3]]
            elements[-1]['properties'].append(prop)

    newelems = {}
    if format == 'ascii':
        body = body.split(newline)
        while len(elements) > 0:
            element = elements.pop(0)
            newelem = {}
            if 'list_dtype' in element['properties'][0]:
                tokens = [np.asarray(ln.split()[1:], dtype=element['properties'][0]['dtype']) for ln in body[:element['len']]]
                newelems[element['name']] = {element['properties'][0]['name']: np.vstack(tokens)}
            else:
                tokens = [ln.split() for ln in body[:element['len']]]
                for idx, prop in enumerate(element['properties']):
                    tmp = [tokens[i][idx] for i in range(len(tokens))]
                    newelem[prop['name']] = np.asarray(tmp, dtype=prop['dtype'])
                newelems[element['name']] = newelem

            body = body[element['len']:]

    v = np.vstack([newelems['vertex'][s] for s in ['x', 'y', 'z']]).T.copy()
    v = np.asarray(v, dtype=np.float64)
    f = newelems['face'].values()[0]

    mesh = Minimal(v=v, f=f)

    return mesh

if __name__ == '__main__':
    pass




#!/usr/bin/env python
# encoding: utf-8
"""
Author(s): Matthew Loper

See LICENCE.txt for licensing and contact information.
"""

import sys
import os
import unittest
import chumpy as ch
from chumpy import Ch
import numpy as np
from util_tests import get_earthmesh

class TestGeometry(unittest.TestCase):

    def setUp(self):
        np.random.seed(0)

    def test_rodrigues(self):
        from geometry import Rodrigues
        rt = np.random.randn(3)
        rt2 = rt + np.random.rand(3)*1e-5
        foo1 = Rodrigues(rt = rt)
        foo2 = Rodrigues(rt = rt2)
        
        empirical = (foo2.r - foo1.r).flatten()
        predicted = foo1.dr_wrt(foo1.rt).dot(rt2-rt)
        
        self.assertTrue(np.max(np.abs(empirical - predicted)) < 1e-10)
        

    def test_vert_normals(self):
        from geometry import VertNormals
        import numpy as np

        mesh = get_earthmesh(np.zeros(3), np.zeros(3))
        v, f = mesh.v*127., mesh.f

        vn1 = VertNormals(f=f, v=v)
        dr_predicted = vn1.dr_wrt(vn1.v).copy()

        eps = .00001 * np.random.randn(v.size).reshape(v.shape)
        v += eps
        vn2 = VertNormals(v=v, f=f)
        empirical_diff = (vn2.r - vn1.r).reshape((-1,3))
        
        predicted_diff = dr_predicted.dot(eps.flatten()).reshape((-1,3))   

        if False:
            print np.max(np.abs(empirical_diff-predicted_diff))
            print empirical_diff[:6]
            print predicted_diff[:6]
        self.assertTrue(np.max(np.abs(empirical_diff-predicted_diff)) < 6e-13)



suite = unittest.TestLoader().loadTestsFromTestCase(TestGeometry)

if __name__ == '__main__':
    unittest.main()
